from __future__ import annotations
import logging
from typing import Any, TypeVar, Optional, List, Dict, TYPE_CHECKING

from pyobs.object import Object

if TYPE_CHECKING:
    from pyobs.robotic import TaskSchedule, TaskRunner, TaskArchive

log = logging.getLogger(__name__)


ProxyClass = TypeVar("ProxyClass")


class Script(Object):
    def __init__(self, configuration: Any, **kwargs: Any):
        """Init Script.

        Args:
            comm: Comm object to use
            observer: Observer to use
        """
        Object.__init__(self, **kwargs)

        # store
        self.exptime_done: float = 0.0
        self.configuration = configuration

    async def can_run(self) -> bool:
        """Whether this config can currently run."""
        raise NotImplementedError

    async def run(
        self,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
    ) -> None:
        """Run script.

        Raises:
            InterruptedError: If interrupted
        """
        raise NotImplementedError

    def get_fits_headers(self, namespaces: Optional[List[str]] = None) -> Dict[str, Any]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """
        return {}


__all__ = ["Script"]



"""
Source Detection
----------------
"""
from .sourcedetection import SourceDetection
from .pysep import SepSourceDetection
from .daophot import DaophotSourceDetection

__all__ = ["SourceDetection", "SepSourceDetection", "DaophotSourceDetection"]



from abc import ABCMeta, abstractmethod
from typing import List, Any

from .interface import Interface
from pyobs.utils.enums import ImageFormat


class IImageFormat(Interface, metaclass=ABCMeta):
    """The module supports different image formats (e.g. INT16, FLOAT32), mainly used by cameras."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_image_format(self, fmt: ImageFormat, **kwargs: Any) -> None:
        """Set the camera image format.

        Args:
            fmt: New image format.

        Raises:
            ValueError: If format could not be set.
        """
        ...

    @abstractmethod
    async def get_image_format(self, **kwargs: Any) -> ImageFormat:
        """Returns the camera image format.

        Returns:
            Current image format.
        """
        ...

    @abstractmethod
    async def list_image_formats(self, **kwargs: Any) -> List[str]:
        """List available image formats.

        Returns:
            List of available image formats.
        """
        ...


__all__ = ["IImageFormat"]



import threading
from typing import Optional

from astropy.time import Time


def check_abort(abort_event: threading.Event, end: Optional[Time] = None) -> None:
    """Throws an exception, if abort_event is set or window has passed

    Args:
        abort_event: Event to check
        end: End of observing window for task

    Raises:
        InterruptedError: if task should be aborted
    """

    # check abort event
    if abort_event.is_set():
        raise InterruptedError

    # check time
    if end is not None and end < Time.now():
        raise InterruptedError



from abc import ABCMeta
from typing import Union, List, Dict, Any
import logging

from pyobs.interfaces import ITelescope, ICamera
from pyobs.mixins.pipeline import PipelineMixin
from pyobs.object import get_object
from pyobs.utils.offsets import ApplyOffsets
from pyobs.modules import Module
from pyobs.images import ImageProcessor
from pyobs.utils import exceptions as exc

log = logging.getLogger(__name__)


class BasePointing(Module, PipelineMixin, metaclass=ABCMeta):
    """Base class for guiding and acquisition modules."""

    __module__ = "pyobs.modules.pointing"

    def __init__(
        self,
        camera: Union[str, ICamera],
        telescope: Union[str, ITelescope],
        pipeline: List[Union[Dict[str, Any], ImageProcessor]],
        apply: Union[Dict[str, Any], ApplyOffsets],
        **kwargs: Any,
    ):
        """Initializes a new base pointing.

        Args:
            telescope: Telescope to use.
            pipeline: Pipeline steps to run on new image. MUST include a step calculating offsets!
            apply: Object that handles applying offsets to telescope.
            log_file: Name of file to write log to.
            log_absolute: Log absolute offsets instead of relative ones to last one.
        """
        Module.__init__(self, **kwargs)
        PipelineMixin.__init__(self, pipeline)

        # store
        self._camera = camera
        self._telescope = telescope

        # apply offsets
        self._apply = get_object(apply, ApplyOffsets)

        # register exceptions
        if isinstance(camera, str):
            exc.register_exception(
                exc.RemoteError, 3, timespan=600, module=camera, callback=self._default_remote_error_callback
            )
        if isinstance(telescope, str):
            exc.register_exception(
                exc.RemoteError, 3, timespan=600, module=telescope, callback=self._default_remote_error_callback
            )

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # check telescope
        try:
            await self.proxy(self._telescope, ITelescope)
        except ValueError:
            log.warning("Given telescope does not exist or is not of correct type at the moment.")

        # check camera
        try:
            await self.proxy(self._camera, ICamera)
        except ValueError:
            log.warning("Given camera does not exist or is not of correct type at the moment.")


__all__ = ["BasePointing"]



from abc import ABCMeta, abstractmethod
from typing import List, Any

from .IMotion import IMotion


class IFilters(IMotion, metaclass=ABCMeta):
    """The module can change filters in a device."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def list_filters(self, **kwargs: Any) -> List[str]:
        """List available filters.

        Returns:
            List of available filters.
        """
        ...

    @abstractmethod
    async def set_filter(self, filter_name: str, **kwargs: Any) -> None:
        """Set the current filter.

        Args:
            filter_name: Name of filter to set.

        Raises:
            ValueError: If an invalid filter was given.
            MoveError: If filter wheel cannot be moved.
        """
        ...

    @abstractmethod
    async def get_filter(self, **kwargs: Any) -> str:
        """Get currently set filter.

        Returns:
            Name of currently set filter.
        """
        ...


__all__ = ["IFilters"]



import logging
from typing import Tuple, Any
import numpy as np

from pyobs.images import Image
from . import Offsets

log = logging.getLogger(__name__)


class FitsHeaderOffsets(Offsets):
    """An offset-calculation method based on fits headers."""

    def __init__(self, target: Tuple[str, str], center: Tuple[str, str] = ("DET-CPX1", "DET-CPX2"), **kwargs: Any):
        """Initializes new fits header offsets."""
        Offsets.__init__(self, **kwargs)

        # store
        self.center = center
        self.target = target

    async def __call__(self, image: Image) -> Image:
        """Processes an image and sets x/y pixel offset to reference in offset attribute.

        Args:
            image: Image to process.

        Returns:
            Original image.

        Raises:
            ValueError: If offset could not be found.
        """

        # get values from header
        target = [image.header[x] for x in self.target]
        center = [image.header[x] for x in self.center]

        # calculate offset
        image.meta["offsets"] = np.subtract(target, center)
        return image


__all__ = ["FitsHeaderOffsets"]



import logging
from typing import Any, Tuple
from astropy.stats import SigmaClip

from pyobs.images.processor import ImageProcessor
from pyobs.images import Image

log = logging.getLogger(__name__)


class RemoveBackground(ImageProcessor):
    """Remove background from image."""

    __module__ = "pyobs.images.processors.misc"

    def __init__(
        self,
        sigma: float = 3.0,
        box_size: Tuple[int, int] = (50, 50),
        filter_size: Tuple[int, int] = (3, 3),
        **kwargs: Any,
    ):
        """Init an image processor that removes background from image.

        Args:
            sigma: Sigma for clipping
            box_size: Box size for bkg estimation.
            filter_size: Size of filter.
        """
        ImageProcessor.__init__(self, **kwargs)

        # store
        self.sigma = sigma
        self.box_size = box_size
        self.filter_size = filter_size

    async def __call__(self, image: Image) -> Image:
        """Remove background from image.

        Args:
            image: Image to remove background from.

        Returns:
            Image without background.
        """
        from photutils.background import Background2D, MedianBackground

        # init objects
        sigma_clip = SigmaClip(sigma=self.sigma)
        bkg_estimator = MedianBackground()

        # calculate background
        bkg = Background2D(
            image.data, self.box_size, filter_size=self.filter_size, sigma_clip=sigma_clip, bkg_estimator=bkg_estimator
        )

        # copy image and remove background
        img = image.copy()
        img.data = img.data - bkg.background
        return img


__all__ = ["RemoveBackground"]



"""
Camera modules.
TODO: write doc
"""
__title__ = "Cameras"

from .basecamera import BaseCamera
from .basevideo import BaseVideo
from .basespectrograph import BaseSpectrograph
from .dummycamera import DummyCamera
from .dummyspectrograph import DummySpectrograph



import logging
from typing import Any
import aiohttp
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
import astropy.units as u

from pyobs.images import Image
from .astrometry import Astrometry


log = logging.getLogger(__name__)


class AstrometryDotNet(Astrometry):
    """Perform astrometry using astrometry.net"""

    __module__ = "pyobs.images.processors.astrometry"

    def __init__(self, url: str, source_count: int = 50, radius: float = 3.0, timeout: int = 10, **kwargs: Any):
        """Init new astronomy.net processor.

        Args:
            url: URL to service.
            source_count: Number of sources to send.
            radius: Radius to search in.
            timeout: Timeout in seconds for call to astrometry web service.
        """
        Astrometry.__init__(self, **kwargs)

        # URL to web-service
        self.url = url
        self.source_count = source_count
        self.radius = radius
        self.timeout = timeout

    async def __call__(self, image: Image) -> Image:
        """Find astrometric solution on given image.

        Writes WCSERR=1 into FITS header on failure.

        Args:
            image: Image to analyse.
        """

        # copy image
        img = image.copy()

        # get catalog
        if img.catalog is None:
            log.warning("No catalog found in image.")
            return image
        cat = img.catalog[["x", "y", "flux"]].to_pandas().dropna()

        # nothing?
        if cat is None or len(cat) < 3:
            log.warning("Not enough sources for astrometry.")
            img.header["WCSERR"] = 1
            return img

        # sort it and take N brightest sources
        cat = cat.sort_values("flux", ascending=False)
        cat = cat[: self.source_count]

        # no CDELT1?
        if "CDELT1" not in img.header:
            log.warning("No CDELT1 found in header.")
            img.header["WCSERR"] = 1
            return img

        # build request data
        scale = abs(img.header["CDELT1"]) * 3600
        data = {
            "ra": img.header["TEL-RA"],
            "dec": img.header["TEL-DEC"],
            "scale_low": scale * 0.9,
            "scale_high": scale * 1.1,
            "radius": self.radius,
            "nx": img.header["NAXIS1"],
            "ny": img.header["NAXIS2"],
            "x": cat["x"].tolist(),
            "y": cat["y"].tolist(),
            "flux": cat["flux"].tolist(),
        }

        # log it
        ra_dec = SkyCoord(ra=data["ra"] * u.deg, dec=data["dec"] * u.deg, frame="icrs")
        cx, cy = img.header["CRPIX1"], img.header["CRPIX2"]
        log.info(
            "Found original RA=%s (%.4f), Dec=%s (%.4f) at pixel %.2f,%.2f.",
            ra_dec.ra.to_string(sep=":", unit=u.hour, pad=True),
            data["ra"],
            ra_dec.dec.to_string(sep=":", unit=u.deg, pad=True),
            data["dec"],
            cx,
            cy,
        )

        # send it
        async with aiohttp.ClientSession() as session:
            async with session.post(self.url, json=data, timeout=self.timeout) as response:
                status_code = response.status
                json = await response.json()

        # success?
        if status_code != 200 or "error" in json:
            # set error
            img.header["WCSERR"] = 1
            if "error" in json:
                # "Could not find WCS file." is just an info, which means that WCS was not successful
                if json["error"] == "Could not find WCS file.":
                    log.info("Could not determine WCS.")
                else:
                    log.warning("Received error from astrometry service: %s", json["error"])
            else:
                log.error("Could not connect to astrometry service.")
            return img

        else:
            # copy keywords
            hdr = json
            header_keywords_to_update = [
                "CTYPE1",
                "CTYPE2",
                "CRPIX1",
                "CRPIX2",
                "CRVAL1",
                "CRVAL2",
                "CD1_1",
                "CD1_2",
                "CD2_1",
                "CD2_2",
            ]
            for keyword in header_keywords_to_update:
                img.header[keyword] = hdr[keyword]

            # astrometry.net gives a CD matrix, so we have to delete the PC matrix and the CDELT* parameters
            for keyword in ["PC1_1", "PC1_2", "PC2_1", "PC2_2", "CDELT1", "CDELT2"]:
                del img.header[keyword]

            # calculate world coordinates for all sources in catalog
            image_wcs = WCS(img.header)
            ras, decs = image_wcs.all_pix2world(img.catalog["x"], img.catalog["y"], 1)

            # set them
            img.catalog["ra"] = ras
            img.catalog["dec"] = decs

            # RA/Dec at center pos
            final_ra, final_dec = image_wcs.all_pix2world(cx, cy, 0)
            ra_dec = SkyCoord(ra=final_ra * u.deg, dec=final_dec * u.deg, frame="icrs")

            # log it
            log.info(
                "Found final RA=%s (%.4f), Dec=%s (%.4f) at pixel %.2f,%.2f.",
                ra_dec.ra.to_string(sep=":", unit=u.hour, pad=True),
                data["ra"],
                ra_dec.dec.to_string(sep=":", unit=u.deg, pad=True),
                data["dec"],
                cx,
                cy,
            )

            # success
            img.header["WCSERR"] = 0

        # finished
        return img


__all__ = ["AstrometryDotNet"]



import asyncio
import logging
from collections import Awaitable
from typing import Union, Tuple, Type, Optional, Any
from astropy.coordinates import SkyCoord
import astropy.units as u

from pyobs.utils import exceptions as exc
from pyobs.modules import Module
from pyobs.interfaces import IPointingAltAz, IPointingRaDec, IReady


log = logging.getLogger(__name__)


async def get_coords(
    obj: Union[IPointingAltAz, IPointingRaDec], mode: Type[Union[IPointingAltAz, IPointingRaDec]]
) -> Tuple[float, float]:
    """Gets coordinates from object

    Args:
        obj: Object to fetch coordinates from.
        mode: IAltAz or IRaDec.

    Returns:
        Return from method call.
    """

    if mode == IPointingAltAz and isinstance(obj, IPointingAltAz):
        return await obj.get_altaz()
    elif mode == IPointingRaDec and isinstance(obj, IPointingRaDec):
        return await obj.get_radec()
    else:
        raise ValueError("Unknown mode.")


def build_skycoord(coord: Tuple[float, float], mode: Type[Union[IPointingAltAz, IPointingRaDec]]) -> SkyCoord:
    """Build SkyCoord from x/y tuple in given mode.

    Args:
        coord: x/y tuple with coordinates in degrees.
        mode: IAltAz or IRaDec.

    Returns:
        SkyCoord with coordinates.
    """

    if mode == IPointingAltAz:
        return SkyCoord(alt=coord[0] * u.deg, az=coord[1] * u.deg, frame="altaz")
    else:
        return SkyCoord(ra=coord[0] * u.deg, dec=coord[1] * u.deg, frame="icrs")


class FollowMixin:
    """Mixin for a device that should follow the motion of another."""

    __module__ = "pyobs.mixins"

    def __init__(
        self,
        device: Optional[str],
        mode: Type[Union[IPointingAltAz, IPointingRaDec]],
        interval: float = 10,
        tolerance: float = 1,
        only_follow_when_ready: bool = True,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the mixin.

        Args:
            device: Name of device to follow
            interval: Interval in seconds between position checks.
            tolerance: Tolerance in degrees between both devices to trigger new movement.
            mode: Set to "altaz" to follow Alt/Az coordinates or "radec" to follow RA/Dec.
            only_follow_when_ready: Only follow if is_ready() is True.
        """

        # store
        self.__follow_device = device
        self.__follow_interval = interval
        self.__follow_tolerance = tolerance
        self.__follow_mode = mode
        self.__follow_only_when_ready = only_follow_when_ready

        # store self for later
        this = self

        # add thread function only, if device is given
        if self.__follow_device is not None:
            if not isinstance(self, Module):
                raise ValueError("This is not a module.")
            self.add_background_task(this.__update_follow)

        # check
        if not isinstance(self, self.__follow_mode):
            raise ValueError("This module is not of given mode %s." % mode)

    @property
    def is_following(self) -> bool:
        """Returns True, if we're following another device."""
        return self.__follow_device is not None

    async def __update_follow(self) -> None:
        """Update function."""

        # store self for later
        this = self
        if not isinstance(self, Module):
            raise ValueError("Not a module.")
        module = self

        # wait a little
        await asyncio.sleep(10)

        # run until closing
        connected = True
        while True:
            # not ready?
            if isinstance(self, IReady):
                if not await self.is_ready() and this.__follow_only_when_ready:
                    await asyncio.sleep(this.__follow_interval)
                    continue

            # get other device
            try:
                device = await module.proxy(this.__follow_device, this.__follow_mode)
            except ValueError:
                # cannot follow, wait a little longer
                log.warning("Cannot follow module, since it is of wrong type.")
                await asyncio.sleep(this.__follow_interval * 10)
                continue

            # get coordinates from other and from myself
            try:
                my_coords = build_skycoord(await get_coords(module, this.__follow_mode), this.__follow_mode)
                xy_coords = await get_coords(device, this.__follow_mode)
                if xy_coords is None:
                    continue
                other_coords = build_skycoord(xy_coords, this.__follow_mode)
                connected = True
            except (ValueError, exc.RemoteError):
                if connected:
                    log.warning("Could not fetch coordinates.")
                connected = False
                await asyncio.sleep(this.__follow_interval * 10.0)
                continue

            # is separation larger than tolerance?
            if my_coords.separation(other_coords).degree > this.__follow_tolerance:
                # move to other
                if this.__follow_mode == IPointingAltAz and isinstance(self, IPointingAltAz):
                    asyncio.create_task(this.__follow__move(self.move_altaz(*xy_coords)))
                elif this.__follow_mode == IPointingRaDec and isinstance(self, IPointingRaDec):
                    asyncio.create_task(this.__follow__move(self.move_radec(*xy_coords)))
                else:
                    raise ValueError("invalid follow mode.")

            # sleep a little
            await asyncio.sleep(this.__follow_interval)

    @staticmethod
    async def __follow__move(awaitable: Awaitable[None]) -> None:
        try:
            await awaitable
        except InterruptedError:
            # ignore interrupts, since probably it was this module who interrupted
            pass


__all__ = ["FollowMixin"]



import asyncio
import io
import logging
from typing import Union, Any, Optional
import numpy as np
from aiohttp import web

from pyobs.interfaces import ICamera, IExposureTime, IWindow
from pyobs.modules import Module
from pyobs.interfaces import IStartStop

log = logging.getLogger(__name__)


class Kiosk(Module, IStartStop):
    """A kiosk mode for a pyobs camera that takes images and published them via HTTP."""

    __module__ = "pyobs.modules.utils"

    def __init__(self, camera: Union[ICamera, str], port: int = 37077, **kwargs: Any):
        """Initializes file cache.

        Args:
            camera: Camera to use for kiosk mode.
            port: Port for HTTP server.
        """
        Module.__init__(self, **kwargs)

        # add thread funcs
        self.add_background_task(self._camera_thread)

        # store stuff
        self._is_listening = False
        self._camera = camera
        self._port = port
        self._exp_time = 2
        self._running = False
        self._image: Optional[bytes] = None

        # create empty image
        from PIL import Image, ImageDraw

        img = Image.new("RGB", (300, 300), color=(0, 0, 0))
        d = ImageDraw.Draw(img)
        d.text((110, 150), "No image taken yet", fill=(255, 255, 255))

        # create image from data array
        with io.BytesIO() as bio:
            img.save(bio, format="jpeg")
            self._empty = bio.getvalue()

        # define web server
        self._app = web.Application()
        self._app.add_routes([web.get("/image.jpg", self.image_handler)])
        self._runner = web.AppRunner(self._app)
        self._site: Optional[web.TCPSite] = None

    async def open(self) -> None:
        """Open server"""
        await Module.open(self)

        # start listening
        log.info("Starting HTTP server on port %d...", self._port)
        await self._runner.setup()
        self._site = web.TCPSite(self._runner, "localhost", self._port)
        await self._site.start()
        self._is_listening = True

    async def close(self) -> None:
        """Close server"""
        await Module.close(self)

        # stop server
        await self._runner.cleanup()

    async def image_handler(self, request: web.Request) -> web.Response:
        """Handles access to /* and returns a specified image.

        Args:
            request: Request to respond to.

        Returns:
            Response containing image.
        """

        # get image
        image = self._empty if self._image is None else self._image

        # send it
        return web.Response(body=image, content_type="image/fits")

    @property
    def opened(self) -> bool:
        """Whether the server is started."""
        return self._is_listening

    async def start(self, **kwargs: Any) -> None:
        """Start kiosk mode."""
        self._running = True

    async def stop(self, **kwargs: Any) -> None:
        """Stop kiosk mode."""
        self._running = False

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether kiosk mode is running."""
        return self._running

    async def _camera_thread(self) -> None:
        """Thread for taking images."""

        # loop until closing
        while True:
            # are we running?
            if not self._running:
                # no, so wait a little and continue
                await asyncio.sleep(1)
                continue

            # get camera
            try:
                camera = await self.proxy(self._camera, ICamera)
            except ValueError:
                await asyncio.sleep(10)
                continue

            # do settings
            if isinstance(camera, IExposureTime):
                # set exposure time
                await camera.set_exposure_time(self._exp_time)
            if isinstance(camera, IWindow):
                # set full frame
                full_frame = await camera.get_full_frame()
                await camera.set_window(*full_frame)

            # do exposure
            filename = await camera.grab_image(False)

            # download image
            try:
                image = await self.vfs.read_image(filename)
            except FileNotFoundError:
                continue

            # convert it to JPEG
            self._image = await asyncio.get_running_loop().run_in_executor(None, image.to_jpeg)

            # adjust exposure time?
            if isinstance(camera, IExposureTime):
                # get max value in image
                max_val = np.max(image.data)

                # adjust
                self._exp_time = self._exp_time / max_val * 40000

                # cut
                self._exp_time = max(self._exp_time, 30)


__all__ = ["Kiosk"]



import logging
from typing import Any

from pyobs.events import NewImageEvent
from pyobs.images.processor import ImageProcessor
from pyobs.images import Image
from pyobs.utils.enums import ImageType
from pyobs.utils.fits import FilenameFormatter

log = logging.getLogger(__name__)


class Broadcast(ImageProcessor):
    """Broadcast image."""

    __module__ = "pyobs.images.processors.misc"

    def __init__(self, filename: str = "/cache/processed_{ORIGNAME}", **kwargs: Any):
        """Init an image processor that broadcasts an image

        Args:
            filename: Filename to broadcast image.
        """
        ImageProcessor.__init__(self, **kwargs)

        # store
        self._formatter = FilenameFormatter(filename)

    async def open(self) -> None:
        """Initialize processor."""
        await ImageProcessor.open(self)

        # register event
        if self.comm is not None:
            await self.comm.register_event(NewImageEvent)

    async def __call__(self, image: Image) -> Image:
        """Broadcast image.

        Args:
            image: Image to broadcast.

        Returns:
            Original image.
        """

        # format filename
        filename = image.format_filename(self._formatter)

        # upload
        await self.vfs.write_image(filename, image)

        # broadcast
        await self.comm.send_event(NewImageEvent(filename, image_type=ImageType(image.header["IMAGETYP"])))

        # finished
        return image


__all__ = ["Broadcast"]



import asyncio
import logging
from typing import Any, Dict

from pyobs.interfaces import IAcquisition
from pyobs.modules import Module
from pyobs.modules import timeout

log = logging.getLogger(__name__)


class DummyAcquisition(Module, IAcquisition):
    """Dummy class for telescope acquisition."""

    __module__ = "pyobs.modules.acquisition"

    def __init__(self, **kwargs: Any):
        """Create a new dummy acquisition."""
        Module.__init__(self, **kwargs)

        # store
        self._is_running = False

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether a service is running."""
        return self._is_running

    @timeout(120)
    async def acquire_target(self, **kwargs: Any) -> Dict[str, Any]:
        """Acquire target at given coordinates.

        If no RA/Dec are given, start from current position. Might not work for some implementations that require
        coordinates.

        Returns:
            A dictionary with entries for datetime, ra, dec, alt, az, and either off_ra, off_dec or off_alt, off_az.

        Raises:
            ValueError: If target could not be acquired.
        """

        try:
            self._is_running = True
            return await self._acquire()
        finally:
            self._is_running = False

    async def _acquire(self) -> Dict[str, Any]:
        """Actually acquire target."""
        log.info("Acquiring target.")
        await asyncio.sleep(5)
        log.info("Finished.")
        return {}


__all__ = ["DummyAcquisition"]



from abc import ABCMeta, abstractmethod
from typing import Optional, Any, List, Dict, Type
from astroplan import ObservingBlock

from pyobs.utils.time import Time
from .task import Task
from pyobs.object import Object


class TaskSchedule(Object, metaclass=ABCMeta):
    def __init__(self, **kwargs: Any):
        Object.__init__(self, **kwargs)

    @abstractmethod
    async def set_schedule(self, blocks: List[ObservingBlock], start_time: Time) -> None:
        """Set the list of scheduled blocks.

        Args:
            blocks: Scheduled blocks.
            start_time: Start time for schedule.
        """
        ...

    @abstractmethod
    async def last_scheduled(self) -> Optional[Time]:
        """Returns time of last scheduler run."""
        ...

    @abstractmethod
    async def get_schedule(self, start_before: Time, end_after: Time, include_running: bool = True) -> Dict[str, Task]:
        """Fetch schedule from portal.

        Args:
            start_before: Task must start before this time.
            end_after: Task must end after this time.
            include_running: Whether to include a currently running task.

        Returns:
            Dictionary with tasks.

        Raises:
            Timeout: If request timed out.
            ValueError: If something goes wrong.
        """
        ...

    @abstractmethod
    def get_task(self, time: Time) -> Optional[Task]:
        """Returns the active task at the given time.

        Args:
            time: Time to return task for.

        Returns:
            Task at the given time.
        """
        ...

    def _create_task(self, klass: Type[Task], **kwargs: Any) -> Task:
        return self.get_object(klass, Task, tasks=self, **kwargs)


__all__ = ["TaskSchedule"]



import logging
from typing import Any

import numpy as np
from astropy.coordinates import EarthLocation

from pyobs.images import Image
from .applyoffsets import ApplyOffsets
from ...interfaces import ITelescope, IOffsetsRaDec

log = logging.getLogger(__name__)


class ApplyRaDecOffsets(ApplyOffsets):
    """Apply offsets from a given image to a given telescope."""

    __module__ = "pyobs.utils.offsets"

    def __init__(self, min_offset: float = 0.5, max_offset: float = 30, **kwargs: Any):
        """Initializes a new ApplyRaDecOffsets.

        Args:
            min_offset: Min offset in arcsec to move.
            max_offset: Max offset in arcsec to move.
        """
        ApplyOffsets.__init__(self, **kwargs)

        # store
        self._min_offset = min_offset
        self._max_offset = max_offset

    async def __call__(self, image: Image, telescope: ITelescope, location: EarthLocation) -> bool:
        """Take the pixel offsets stored in the meta data of the image and apply them to the given telescope.

        Args:
            image: Image to take offsets and WCS from.
            telescope: Telescope to apply offsets to.
            location: Observer location on Earth.

        Returns:
            Whether offsets have been applied successfully.
        """

        # telescope must be of type IRaDecOffsets
        tel = telescope
        if not isinstance(telescope, IOffsetsRaDec):
            log.error("Given telescope cannot handle RA/Dec offsets.")
            return False

        # get RA/Dec coordinates of center and center+offsets
        try:
            radec_center, radec_target = self._get_radec_center_target(image, location)
        except ValueError:
            log.warning("Could not get offsets from image meta.")
            return False

        # get offset
        dra, ddec = radec_center.spherical_offsets_to(radec_target)
        log.info('Transformed to RA/Dec shift of dRA=%.2f", dDec=%.2f".', dra.arcsec, ddec.arcsec)

        # get current offset
        cur_dra, cur_ddec = await telescope.get_offsets_radec()

        # log it
        await self._log_offset(tel, "dra", cur_dra, dra.degree, "ddec", cur_ddec, ddec.degree)

        # too large or too small?
        diff = np.sqrt(dra.arcsec ** 2.0 + ddec.arcsec ** 2)
        if diff < self._min_offset:
            log.warning("Shift too small, skipping auto-guiding for now...")
            return False
        if diff > self._max_offset:
            log.warning("Shift too large, skipping auto-guiding for now...")
            return False

        # move offset
        log.info("Offsetting telescope...")
        await telescope.set_offsets_radec(float(cur_dra + dra.degree), float(cur_ddec + ddec.degree))
        return True


__all__ = ["ApplyRaDecOffsets"]



import logging
from typing import Tuple, Any, Optional, List
import numpy as np
import numpy.typing as npt
from scipy.interpolate import UnivariateSpline
from scipy.optimize import fmin
import re

from pyobs.images import Image
from pyobs.utils.pid import PID
from pyobs.images.meta import PixelOffsets
from .offsets import Offsets

log = logging.getLogger(__name__)


class ProjectedOffsets(Offsets):
    """An auto-guiding system based on comparing collapsed images along the x&y axes with a reference image."""

    __module__ = "pyobs.images.processors.offsets"

    def __init__(self, **kwargs: Any):
        """Initializes a new auto guiding system."""
        Offsets.__init__(self, **kwargs)

        # init
        self._ref_image: Optional[Tuple[npt.NDArray[float], npt.NDArray[float]]] = None
        self._pid_ra = None
        self._pid_dec = None

    async def reset(self) -> None:
        """Resets guiding."""
        log.info("Reset auto-guiding.")
        self._ref_image = None

    async def __call__(self, image: Image) -> Image:
        """Processes an image and sets x/y pixel offset to reference in offset attribute.

        Args:
            image: Image to process.

        Returns:
            Original image.

        Raises:
            ValueError: If offset could not be found.
        """

        # no reference image?
        if self._ref_image is None:
            log.info("Initialising auto-guiding with new image...")
            self._ref_image = self._process(image)
            self._init_pid()
            self.offset = (0, 0)
            return image

        # process it
        log.info("Perform auto-guiding on new image...")
        sum_x, sum_y = self._process(image)

        # find peaks
        dx = self._correlate(sum_x, self._ref_image[0])
        dy = self._correlate(sum_y, self._ref_image[1])
        if dx is None or dy is None:
            log.warning("Could not correlate peaks.")
            return image

        # set it
        image.set_meta(PixelOffsets(dx, dy))
        return image

    def _process(self, image: Image) -> Tuple[npt.NDArray[float], npt.NDArray[float]]:
        """Project image along x and y axes and return results.

        Args:
            image: Image to process.

        Returns:
            Projected images.
        """

        # get image data and header
        data, hdr = image.data, image.header

        # no data?
        if data is None:
            raise ValueError("Image contains no data.")

        # trimsec
        if "TRIMSEC" in hdr:
            m = re.match(r"\[([0-9]+):([0-9]+),([0-9]+):([0-9]+)\]", hdr["TRIMSEC"])
            if m is None:
                raise ValueError("Invalid trimsec.")
            x0, x1, y0, y1 = [int(f) for f in m.groups()]
            data = data[y0 - 1 : y1, x0 - 1 : x1]

        # collapse
        sum_x = np.nansum(data, 0)
        sum_y = np.nansum(data, 1)

        # sky subtraction
        return self._subtract_sky(sum_x), self._subtract_sky(sum_y)

    @staticmethod
    def _gaussian(pars: List[float], x: npt.NDArray[float]) -> npt.NDArray[float]:
        a = pars[0]
        x0 = pars[1]
        sigma = pars[2]
        return a * np.exp(-((x - x0) ** 2) / (2.0 * sigma ** 2))

    @staticmethod
    def _gaussian_fit(pars: List[float], y: npt.NDArray[float], x: npt.NDArray[float]) -> float:
        err = y - ProjectedOffsets._gaussian(pars, x)
        return (err * err).sum()

    @staticmethod
    def _correlate(data1: npt.NDArray[float], data2: npt.NDArray[float], fit_width: int = 10) -> Optional[float]:
        # do cross-correlation
        corr = np.correlate(data1, data2, "full")

        # find index of maximum
        i_max = np.argmax(corr)
        centre = i_max - data1.size + 1

        # cut window
        x = np.linspace(centre - fit_width, centre + fit_width, 2 * fit_width + 1)
        y = corr[i_max - fit_width : i_max + fit_width + 1]

        # moment calculation for initial guesses
        total = float(y.sum())
        m = (x * y).sum() / total
        m2 = (x * x * y).sum() / total - m ** 2

        # initial guess
        guesses = [np.max(y), m, m2]

        # perform fit
        result = fmin(ProjectedOffsets._gaussian_fit, guesses, args=(y, x), disp=False)

        # sanity check and finish up
        shift = float(result[1])
        if shift < centre - fit_width or shift > centre + fit_width:
            return None
        return shift

    def _init_pid(self) -> None:
        # init pids
        Kp = 0.2
        Ki = 0.16
        Kd = 0.83

        # reset
        self._pid_ra = PID(Kp, Ki, Kd)
        self._pid_dec = PID(Kp, Ki, Kd)

    @staticmethod
    def _subtract_sky(data: npt.NDArray[float], frac: float = 0.15, sbin: int = 10) -> npt.NDArray[float]:
        # find continuum for every of the sbin bins
        bins = np.zeros((sbin,))
        binxs = np.zeros((sbin,))
        x = list(range(len(data)))
        w1 = 0
        w2 = float(len(x)) / sbin
        for i in range(sbin):
            # sort data in range
            bindata = list(reversed(sorted(data[int(w1) : int(w2)])))
            # calculate median and set wavelength
            bins[i] = np.median(bindata[int(-frac * len(bindata)) : -1])
            binxs[i] = np.mean(x[int(w1) : int(w2)])
            # reset ranges
            w1 = w2
            w2 += float(len(x)) / sbin
            # check for last bin
            if i == sbin - 1:
                w2 = len(x)

        # fit it
        w = np.where(~np.isnan(bins))
        ip = UnivariateSpline(binxs[w], bins[w])
        cont = ip(x)

        # return continuum
        return data - cont


__all__ = ["ProjectedOffsets"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface
from pyobs.utils.enums import ImageType


class IImageType(Interface, metaclass=ABCMeta):
    """The module supports different image types (e.g. object, bias, dark, etc), mainly used by cameras."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_image_type(self, image_type: ImageType, **kwargs: Any) -> None:
        """Set the image type.

        Args:
            image_type: New image type.
        """
        ...

    @abstractmethod
    async def get_image_type(self, **kwargs: Any) -> ImageType:
        """Returns the current image type.

        Returns:
            Current image type.
        """
        ...


__all__ = ["IImageType"]



import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Optional

from pyobs.images import Image
from pyobs.images.meta.exptime import ExpTime
from pyobs.images.processor import ImageProcessor


log = logging.getLogger(__name__)


class ExpTimeEstimator(ImageProcessor, metaclass=ABCMeta):
    """Estimate exposure time."""

    __module__ = "pyobs.images.processors.exptime"

    def __init__(self, min_exp_time: float = 0.0, max_exp_time: Optional[float] = None, **kwargs: Any):
        """Init new exposure time estimator."""
        ImageProcessor.__init__(self, **kwargs)
        self._min_exp_time = min_exp_time
        self._max_exp_time = max_exp_time

    @abstractmethod
    async def __call__(self, image: Image) -> Image:
        """Processes an image and stores new exposure time in exp_time attribute.

        Args:
            image: Image to process.

        Returns:
            Original image.
        """
        ...

    def _set_exp_time(self, image: Image, exp_time: float) -> None:
        """Internal setter for exposure time."""

        # min exp time
        if self._min_exp_time is not None and exp_time < self._min_exp_time:
            exp_time = self._min_exp_time

        # max exp time
        if self._max_exp_time is not None and exp_time > self._max_exp_time:
            exp_time = self._max_exp_time

        # set it
        image.set_meta(ExpTime(exp_time))


__all__ = ["ExpTimeEstimator"]



import os
from typing import Optional
import logging
import aiohttp

from .httpfile import HttpFile


log = logging.getLogger(__name__)


class ArchiveFile(HttpFile):
    """Wraps a file in an archive. To be used in combination with pyobs-archive."""

    __module__ = "pyobs.vfs"

    def __init__(self, name: str, url: str, mode: str = "w", token: Optional[str] = None):
        """Creates a new archive file.

        Args:
            name: Name of file.
            mode: Open mode (r/w).
            url: Archive url url.
            token: Authorization token.
        """

        # init
        HttpFile.__init__(self, name, mode, upload="")

        # only allow write access for now
        if "w" not in mode:
            raise ValueError("Only write operations allowed.")

        # store
        self._url = url + ("/" if not url.endswith("/") else "")
        self._headers = {"Authorization": "Token " + token} if token is not None else {}

    async def _upload(self) -> None:
        """If in write mode, actually send the file to the archive."""

        # open session
        async with aiohttp.ClientSession() as session:
            # do some initial GET request for getting the csrftoken
            async with session.get(self._url, headers=self._headers) as response:
                token = response.cookies["csrftoken"].value

            # define list of files and url
            url = self._url + "frames/create/"
            data = aiohttp.FormData()
            data.add_field("csrfmiddlewaretoken", token)
            data.add_field("file", self._buffer, filename=os.path.basename(self.filename))

            # send data and return image ID
            async with session.post(url, auth=self._auth, data=data, timeout=10, headers=self._headers) as response:
                # success, if status code is 200
                if response.status != 200:
                    raise ValueError("Cannot write file, received status_code %d." % response.status)

                # check json
                json = await response.json()
                if "created" not in json or json["created"] == 0:
                    if "errors" in json:
                        raise ValueError("Could not create file in archive: " + str(json["errors"]))
                    else:
                        raise ValueError("Could not create file in archive.")


__all__ = ["ArchiveFile"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .ITemperatures import ITemperatures


class ICooling(ITemperatures, metaclass=ABCMeta):
    """The module can control the cooling of a device."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_cooling_status(self, **kwargs: Any) -> Tuple[bool, float, float]:
        """Returns the current status for the cooling.

        Returns:
            (tuple): Tuple containing:
                Enabled:  Whether the cooling is enabled
                SetPoint: Setpoint for the cooling in celsius.
                Power:    Current cooling power in percent or None.
        """
        ...

    @abstractmethod
    async def set_cooling(self, enabled: bool, setpoint: float, **kwargs: Any) -> None:
        """Enables/disables cooling and sets setpoint.

        Args:
            enabled: Enable or disable cooling.
            setpoint: Setpoint in celsius for the cooling.

        Raises:
            ValueError: If cooling could not be set.
        """
        ...


__all__ = ["ICooling"]



"""
Using interface, a :class:`~pyobs.modules.Module` signals another one, what functionality it provides for remote
procedure calls. The base class for all interfaces in *pyobs* is:

.. autoclass:: pyobs.interfaces.Interface
   :members:

Modules need to implement the required interfaces. For instance, if a module operates a camera, it probably should
implement :class:`~pyobs.interfaces.ICamera`.
"""
__title__ = "Interfaces"

from .IAbortable import IAbortable
from .IAcquisition import IAcquisition
from .ILatLon import ILatLon
from .IAutoFocus import IAutoFocus
from .IAutoGuiding import IAutoGuiding
from .IAutonomous import IAutonomous
from .IBinning import IBinning
from .ICalibrate import ICalibrate
from .ICamera import ICamera
from .IConfig import IConfig
from .ICooling import ICooling
from .IDome import IDome
from .IExposureTime import IExposureTime
from .IFilters import IFilters
from .IFitsHeaderAfter import IFitsHeaderAfter
from .IFitsHeaderBefore import IFitsHeaderBefore
from .IFlatField import IFlatField
from .IFocusModel import IFocusModel
from .IFocuser import IFocuser
from .IGain import IGain
from .IImageFormat import IImageFormat
from .IImageGrabber import IImageGrabber
from .IImageType import IImageType
from .IModule import IModule
from .IMotion import IMotion
from .IPointingAltAz import IPointingAltAz
from .IPointingHGS import IPointingHGS
from .IPointingRaDec import IPointingRaDec
from .IPointingSeries import IPointingSeries
from .IReady import IReady
from .IRoof import IRoof
from .IRotation import IRotation
from .IRunnable import IRunnable
from .IRunning import IRunning
from .IScriptRunner import IScriptRunner
from .IStartStop import IStartStop
from .ISyncTarget import ISyncTarget
from .ITelescope import ITelescope
from .ITemperatures import ITemperatures
from .IOffsetsAltAz import IOffsetsAltAz
from .IOffsetsRaDec import IOffsetsRaDec
from .ISpectrograph import ISpectrograph
from .IVideo import IVideo
from .IWeather import IWeather
from .IWindow import IWindow
from .interface import Interface



import logging
from typing import Union, List, Dict, Any, Optional
from pyobs.images import ImageProcessor, Image

from pyobs.object import Object

log = logging.getLogger(__name__)


class PipelineMixin:
    """Mixin for a module that needs to implement an image pipeline."""

    __module__ = "pyobs.mixins"

    def __init__(self, steps: Optional[List[Union[Dict[str, Any], ImageProcessor]]] = None, **kwargs: Any):
        """Initializes the mixin.

        Args:
            steps: Pipeline steps to run on images.
        """

        # store
        if isinstance(self, Object):
            steps = [] if steps is None else steps
            self.__pipeline_steps = [self.add_child_object(step, ImageProcessor) for step in steps]

        else:
            raise ValueError("This class is no Object.")

    def reset_pipeline(self) -> None:
        """Resets all previous state of the involved image processors."""
        for step in self.__pipeline_steps:
            step.reset()

    async def run_pipeline(self, image: Image) -> Image:
        """Run the pipeline on the given image.

        Args:
            image: Image to run pipeline on.

        Returns:
            Image after pipeline run.
        """

        # loop steps
        for step in self.__pipeline_steps:
            try:
                image = await step(image)
            except Exception as e:
                log.exception(f"Could not run pipeline step {step.__class__.__name__}: {e}")

        # finished
        return image


__all__ = ["PipelineMixin"]



import asyncio
from functools import partial
from typing import Tuple, TYPE_CHECKING, Any, Optional
from astropy.table import Table
import logging
import numpy as np
import numpy.typing as npt
import pandas as pd

from .sourcedetection import SourceDetection
from pyobs.images import Image

if TYPE_CHECKING:
    from sep import Background

log = logging.getLogger(__name__)


class SepSourceDetection(SourceDetection):
    """Detect sources using SEP."""

    __module__ = "pyobs.images.processors.detection"

    def __init__(
        self,
        threshold: float = 1.5,
        minarea: int = 5,
        deblend_nthresh: int = 32,
        deblend_cont: float = 0.005,
        clean: bool = True,
        clean_param: float = 1.0,
        **kwargs: Any,
    ):
        """Initializes a wrapper for SEP. See its documentation for details.

        Highly inspired by LCO's wrapper for SEP, see:
        https://github.com/LCOGT/banzai/blob/master/banzai/photometry.py

        Args:
            threshold: Threshold pixel value for detection.
            minarea: Minimum number of pixels required for detection.
            deblend_nthresh: Number of thresholds used for object deblending.
            deblend_cont: Minimum contrast ratio used for object deblending.
            clean: Perform cleaning?
            clean_param: Cleaning parameter (see SExtractor manual).
        """
        SourceDetection.__init__(self, **kwargs)

        # store
        self.threshold = threshold
        self.minarea = minarea
        self.deblend_nthresh = deblend_nthresh
        self.deblend_cont = deblend_cont
        self.clean = clean
        self.clean_param = clean_param

    async def __call__(self, image: Image) -> Image:
        """Find stars in given image and append catalog.

        Args:
            image: Image to find stars in.

        Returns:
            Image with attached catalog.
        """
        import sep

        loop = asyncio.get_running_loop()

        # got data?
        if image.data is None:
            log.warning("No data found in image.")
            return image

        # no mask?
        mask = image.mask if image.mask is not None else np.zeros(image.data.shape, dtype=bool)

        # remove background
        data, bkg = SepSourceDetection.remove_background(image.data, mask)

        # extract sources
        sources = await loop.run_in_executor(
            None,
            partial(
                sep.extract,
                data,
                self.threshold,
                err=bkg.globalrms,
                minarea=self.minarea,
                deblend_nthresh=self.deblend_nthresh,
                deblend_cont=self.deblend_cont,
                clean=self.clean,
                clean_param=self.clean_param,
                mask=image.mask,
            ),
        )

        # convert to astropy table
        sources = pd.DataFrame(sources)

        # only keep sources with detection flag < 8
        sources = sources[sources["flag"] < 8]
        x, y = sources["x"], sources["y"]

        # Calculate the ellipticity
        sources["ellipticity"] = 1.0 - (sources["b"] / sources["a"])

        # calculate the FWHMs of the stars
        fwhm = 2.0 * (np.log(2) * (sources["a"] ** 2.0 + sources["b"] ** 2.0)) ** 0.5
        sources["fwhm"] = fwhm

        # clip theta to [-pi/2,pi/2]
        sources["theta"] = sources["theta"].clip(lower=np.pi / 2, upper=np.pi / 2)

        # Kron radius
        kronrad, krflag = sep.kron_radius(data, x, y, sources["a"], sources["b"], sources["theta"], 6.0)
        sources["flag"] |= krflag
        sources["kronrad"] = kronrad

        # equivalent of FLUX_AUTO
        gain = image.header["DET-GAIN"] if "DET-GAIN" in image.header else None
        flux, fluxerr, flag = await loop.run_in_executor(
            None,
            partial(
                sep.sum_ellipse,
                data,
                x,
                y,
                sources["a"],
                sources["b"],
                sources["theta"],
                2.5 * kronrad,
                subpix=5,
                mask=image.mask,
                gain=gain,
            ),
        )
        sources["flag"] |= flag
        sources["flux"] = flux

        # radii at 0.25, 0.5, and 0.75 flux
        flux_radii, flag = sep.flux_radius(
            data, x, y, 6.0 * sources["a"], [0.25, 0.5, 0.75], normflux=sources["flux"], subpix=5
        )
        sources["flag"] |= flag
        sources["fluxrad25"] = flux_radii[:, 0]
        sources["fluxrad50"] = flux_radii[:, 1]
        sources["fluxrad75"] = flux_radii[:, 2]

        # xwin/ywin
        sig = 2.0 / 2.35 * sources["fluxrad50"]
        xwin, ywin, flag = sep.winpos(data, x, y, sig)
        sources["flag"] |= flag
        sources["xwin"] = xwin
        sources["ywin"] = ywin

        # theta in degrees
        sources["theta"] = np.degrees(sources["theta"])

        # only keep sources with detection flag < 8
        sources = sources[sources["flag"] < 8]

        # match fits conventions
        sources["x"] += 1
        sources["y"] += 1

        # pick columns for catalog
        cat = sources[
            [
                "x",
                "y",
                "peak",
                "flux",
                "fwhm",
                "a",
                "b",
                "theta",
                "ellipticity",
                "tnpix",
                "kronrad",
                "fluxrad25",
                "fluxrad50",
                "fluxrad75",
                "xwin",
                "ywin",
            ]
        ]

        # copy image, set catalog and return it
        img = image.copy()
        img.catalog = Table.from_pandas(cat)
        return img

    @staticmethod
    def remove_background(
        data: npt.NDArray[float], mask: Optional[npt.NDArray[float]] = None
    ) -> Tuple[npt.NDArray[float], "Background"]:
        """Remove background from image in data.

        Args:
            data: Data to remove background from.
            mask: Mask to use for estimating background.

        Returns:
            Image without background.
        """
        import sep

        # get data and make it continuous
        d = data.astype(float)

        # estimate background, probably we need to byte swap
        try:
            bkg = sep.Background(d, mask=mask, bw=32, bh=32, fw=3, fh=3)
        except ValueError as e:
            d = d.byteswap(True).newbyteorder()
            bkg = sep.Background(d, mask=mask, bw=32, bh=32, fw=3, fh=3)

        # subtract it
        bkg.subfrom(d)

        # return data without background and background
        return d, bkg


__all__ = ["SepSourceDetection"]



from typing import List, Optional, Dict, Any, TYPE_CHECKING

from pyobs.utils.enums import ImageType
from pyobs.utils.time import Time

if TYPE_CHECKING:
    from pyobs.images import Image


class FrameInfo:
    """Base class for frame infos."""

    def __init__(self) -> None:
        self.id = None
        self.filename = None
        self.filter_name = None
        self.binning = None
        self.dateobs = None


class Archive:
    """Base class for image archives."""

    __module__ = "pyobs.utils.archive"

    async def list_options(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> Dict[str, List[Any]]:
        raise NotImplementedError

    async def list_frames(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> List[FrameInfo]:
        raise NotImplementedError

    async def download_frames(self, frames: List[FrameInfo]) -> List["Image"]:
        raise NotImplementedError

    async def upload_frames(self, frames: List["Image"]) -> None:
        raise NotImplementedError


__all__ = ["FrameInfo", "Archive"]



"""
Modules for performing flatfields.
TODO: write doc
"""
__title__ = "Flatfielding"

from .flatfield import FlatField
from .pointing import FlatFieldPointing
from .scheduler import FlatFieldScheduler



import logging
import asyncio
from enum import Enum
from typing import Union, Tuple, List, Optional, Any, Dict

from pyobs.events import BadWeatherEvent, RoofClosingEvent, Event
from pyobs.interfaces import IFlatField, IFilters, IBinning, ITelescope, ICamera
from pyobs.modules import Module
from pyobs.modules import timeout
from pyobs.utils.enums import MotionStatus
from pyobs.utils.publisher import CsvPublisher
from pyobs.utils.skyflats import FlatFielder
from pyobs.utils import exceptions as exc

log = logging.getLogger(__name__)


class FlatField(Module, IFlatField, IBinning, IFilters):
    """Module for auto-focusing a telescope."""

    __module__ = "pyobs.modules.flatfield"

    class Twilight(Enum):
        DUSK = "dusk"
        DAWN = "dawn"

    class State(Enum):
        INIT = "init"
        WAITING = "waiting"
        TESTING = "testing"
        RUNNING = "running"
        FINISHED = "finished"

    def __init__(
        self,
        telescope: Union[str, ITelescope],
        camera: Union[str, ICamera],
        flat_fielder: Optional[Union[Dict[str, Any], FlatFielder]],
        filters: Optional[Union[str, IFilters]] = None,
        log_file: Optional[str] = None,
        **kwargs: Any,
    ):
        """Initialize a new flat fielder.

        Args:
            telescope: Name of ITelescope.
            camera: Name of ICamera.
            flat_fielder: Flat field object to use.
            filters: Name of IFilters, if any.
            log_file: Name of file to store flat field log in.
        """
        Module.__init__(self, **kwargs)

        # store telescope, camera, and filters
        self._telescope = telescope
        self._camera = camera
        self._filter_wheel = filters
        self._abort = asyncio.Event()
        self._running = False

        # flat fielder
        self._flat_fielder = self.get_object(flat_fielder, FlatFielder, callback=self.callback)

        # init log file
        self._publisher = None if log_file is None else CsvPublisher(log_file)

        # init binning and filter
        self._binning = (1, 1)
        self._filter: Optional[str] = None

        # need to add IFilters interface?
        if self._filter_wheel is not None:
            # check filters
            if not self._flat_fielder.has_filters:
                raise ValueError("Filter wheel module given in config, but no filters in functions.")

            # add it
            # self.__class__ = type('FlatFieldFilter', (FlatField, IFilters), {})

        # register exceptions
        if isinstance(camera, str):
            exc.register_exception(
                exc.RemoteError, 3, timespan=600, module=camera, callback=self._default_remote_error_callback
            )
        if isinstance(telescope, str):
            exc.register_exception(
                exc.RemoteError, 3, timespan=600, module=telescope, callback=self._default_remote_error_callback
            )
        if isinstance(filters, str):
            exc.register_exception(
                exc.RemoteError, 3, timespan=600, module=filters, callback=self._default_remote_error_callback
            )

    async def open(self) -> None:
        """Open module"""
        await Module.open(self)

        # check telescope, camera, and filters
        try:
            await self.proxy(self._telescope, ITelescope)
            await self.proxy(self._camera, ICamera)
            await self.proxy(self._filter_wheel, IFilters)
        except ValueError:
            log.warning("Either telescope, camera or filters do not exist or are not of correct type at the moment.")

            # subscribe to events
            if self.comm:
                await self.comm.register_event(BadWeatherEvent, self._abort_weather)
                await self.comm.register_event(RoofClosingEvent, self._abort_weather)

    async def close(self) -> None:
        """Close module."""
        await Module.close(self)
        self._abort.set()

    async def callback(
        self, datetime: str, solalt: float, exptime: float, counts: float, filter_name: str, binning: Tuple[int, int]
    ) -> None:
        """Callback for flat-field class to call with statistics."""
        # write log
        if self._publisher is not None:
            await self._publisher(
                datetime=datetime, solalt=solalt, exptime=exptime, counts=counts, filter=filter_name, binning=binning[0]
            )

    async def list_binnings(self, **kwargs: Any) -> List[Tuple[int, int]]:
        """List available binnings.

        Returns:
            List of available binnings as (x, y) tuples.
        """
        proxy = await self.proxy(self._camera, IBinning)
        return await proxy.list_binnings()

    async def set_binning(self, x: int, y: int, **kwargs: Any) -> None:
        """Set the camera binning.

        Args:
            x: X binning.
            y: Y binning.

        Raises:
            ValueError: If binning could not be set.
        """
        self._binning = (x, y)

    async def get_binning(self, **kwargs: Any) -> Tuple[int, int]:
        """Returns the camera binning.

        Returns:
            Tuple with x and y.
        """
        return self._binning

    async def list_filters(self, **kwargs: Any) -> List[str]:
        """List available filters.

        Returns:
            List of available filters.
        """
        proxy = await self.proxy(self._filter_wheel, IFilters)
        return await proxy.list_filters()

    async def set_filter(self, filter_name: str, **kwargs: Any) -> None:
        """Set the current filter.

        Args:
            filter_name: Name of filter to set.

        Raises:
            ValueError: If binning could not be set.
        """
        self._filter = filter_name

    async def get_filter(self, **kwargs: Any) -> str:
        """Get currently set filter.

        Returns:
            Name of currently set filter.
        """
        return "" if self._filter is None else self._filter

    @timeout(3600)
    async def flat_field(self, count: int = 20, **kwargs: Any) -> Tuple[int, float]:
        """Do a series of flat fields in the given filter.

        Args:
            count: Number of images to take

        Returns:
            Number of images actually taken and total exposure time in ms
        """

        # check
        if self._running:
            raise ValueError("Already running.")
        self._running = True

        try:
            # start
            log.info("Performing flat fielding...")
            self._abort = asyncio.Event()

            # get telescope
            log.info("Getting proxy for telescope...")
            telescope = await self.proxy(self._telescope, ITelescope)

            # get camera
            log.info("Getting proxy for camera...")
            camera = await self.proxy(self._camera, ICamera)

            # get filter wheel
            filters: Optional[IFilters] = None
            if self._filter_wheel is not None:
                log.info("Getting proxy for filter wheel...")
                filters = await self.proxy(self._filter_wheel, IFilters)

            # reset
            await self._flat_fielder.reset()

            # run until state is finished or we aborted
            state = None
            while state != FlatFielder.State.FINISHED:
                # can we run?
                if not await telescope.is_ready():
                    log.error("Telescope not in valid state, aborting...")
                    return self._flat_fielder.image_count, self._flat_fielder.total_exptime
                if self._abort.is_set():
                    log.warning("Aborting flat-fielding...")
                    return self._flat_fielder.image_count, self._flat_fielder.total_exptime

                # do step
                state = await self._flat_fielder(
                    telescope, camera, count=count, binning=self._binning, filters=filters, filter_name=self._filter
                )

            # stop telescope
            log.info("Stopping telescope...")
            await telescope.stop_motion()
            log.info("Flat-fielding finished.")

            # return number of taken images
            return int(self._flat_fielder.image_count), float(self._flat_fielder.total_exptime)

        finally:
            self._running = False

    @timeout(20)
    async def abort(self, **kwargs: Any) -> None:
        """Abort current actions."""
        self._abort.set()

    async def _abort_weather(self, event: Event, sender: str, **kwargs: Any) -> bool:
        """Abort on bad weather."""
        await self.abort()
        return True

    async def init(self, **kwargs: Any) -> None:
        pass

    async def park(self, **kwargs: Any) -> None:
        pass

    async def get_motion_status(self, device: Optional[str] = None, **kwargs: Any) -> MotionStatus:
        return MotionStatus.IDLE

    async def stop_motion(self, device: Optional[str] = None, **kwargs: Any) -> None:
        pass

    async def is_ready(self, **kwargs: Any) -> bool:
        return True


__all__ = ["FlatField"]



from typing import Tuple, Dict, List, Any
import numpy as np
import logging
from pyobs.object import get_object

from pyobs.images.processors.detection import SourceDetection
from pyobs.utils.focusseries.base import FocusSeries
from pyobs.utils.curvefit import fit_hyperbola
from pyobs.images import Image


log = logging.getLogger(__name__)


class PhotometryFocusSeries(FocusSeries):
    """Focus series based on source detection."""

    __module__ = "pyobs.utils.focusseries"

    def __init__(self, source_detection: SourceDetection, radius_column: str = "radius", **kwargs: Any):
        """Initialize a new projection focus series.

        Args:
            source_detection: Photometry to use for estimating PSF sizes
        """

        # stuff
        self._source_detection: SourceDetection = get_object(source_detection, SourceDetection)
        self._radius_col = radius_column
        self._data: List[Dict[str, float]] = []

    def reset(self) -> None:
        """Reset focus series."""
        self._data = []

    def analyse_image(self, image: Image, focus_value: float) -> None:
        """Analyse given image.

        Args:
            image: Image to analyse
            focus_value: Value to fit along, e.g. focus value or its offset
        """

        # do photometry
        self._source_detection(image)

        # filter
        sources = image.catalog
        if sources is None:
            return
        sources = sources[sources["ellipticity"] < 0.1]
        sources = sources[sources["peak"] > 1000]
        sources = sources[sources["radius"] > 0]

        # calculate median radius
        radius = np.median(sources["radius"])
        radius_err = np.std(sources["radius"])

        # log it
        log.info("Found median radius of %.1f+-%.1f.", radius, radius_err)

        # add to list
        self._data.append({"focus": focus_value, "r": radius, "rerr": radius_err})

    def fit_focus(self) -> Tuple[float, float]:
        """Fit focus from analysed images

        Returns:
            Tuple of new focus and its error
        """

        # get data
        focus = [d["focus"] for d in self._data]
        r = [d["r"] for d in self._data]
        rerr = [d["rerr"] for d in self._data]

        # fit focus
        try:
            foc, err = fit_hyperbola(focus, r, rerr)
        except (RuntimeError, RuntimeWarning):
            raise ValueError("Could not find best focus.")

        # get min and max foci
        min_focus = np.min(focus)
        max_focus = np.max(focus)
        if foc < min_focus or foc > max_focus:
            raise ValueError("New focus out of bounds: {0:.3f}+-{1:.3f}mm.".format(foc, err))

        # return it
        return float(foc), float(err)


__all__ = ["PhotometryFocusSeries"]



import logging
import asyncio
from typing import Any

from pyobs.events import NewImageEvent, Event
from pyobs.images import Image
from ._baseguiding import BaseGuiding
from ...utils.parallel import event_wait

log = logging.getLogger(__name__)


class ScienceFrameAutoGuiding(BaseGuiding):
    """An auto-guiding system based on comparing collapsed images along the x&y axes with a reference image."""

    __module__ = "pyobs.modules.guiding"

    def __init__(self, **kwargs: Any):
        """Initializes a new science frame auto guiding system."""
        BaseGuiding.__init__(self, **kwargs)

        # add thread func
        self.add_background_task(self._auto_guiding, True)

        # variables
        self._next_image: asyncio.Queue[Image] = asyncio.Queue()

    async def open(self) -> None:
        """Open module."""
        await BaseGuiding.open(self)

        # subscribe to channel with new images
        log.info("Subscribing to new image events...")
        await self.comm.register_event(NewImageEvent, self.add_image)

    def set_exposure_time(self, exposure_time: float, **kwargs: Any):
        """Set the exposure time for the auto-guider.

        Args:
            exposure_time: Exposure time in secs.
        """
        raise NotImplementedError

    async def add_image(self, event: Event, sender: str, **kwargs: Any) -> bool:
        """Processes an image asynchronously, returns immediately.

        Args:
            event: Event for new image.
            sender: Name of sender.
        """

        # did it come from correct camera and are we enabled?
        if sender != self._camera or not self._enabled or not isinstance(event, NewImageEvent):
            return False
        log.info("Received new image.")

        # download image
        image = await self.vfs.read_image(event.filename)

        # we only accept OBJECT images
        if image.header["IMAGETYP"] != "object":
            return False

        # do we have a filename in here already?
        if not self._next_image.empty():
            log.warning("Last image still being processed by auto-guiding, skipping new one.")
            return False

        # store it
        await self._next_image.put(image)
        return True

    async def _auto_guiding(self) -> None:
        """the thread function for processing the images"""

        # run until closed
        while True:
            # get next image to process
            image = await self._next_image.get()

            # process it
            await self._process_image(image)

            # wait for next image
            await asyncio.sleep(1)


__all__ = ["ScienceFrameAutoGuiding"]



"""
TODO: write doc
"""
__title__ = "Enumerations"

from enum import Enum


class ModuleState(Enum):
    """Enumerator for module states.

    Attributes:
        CLOSED: Module is closed.
        READY: Module is ready.
        ERROR: Module has an error.
        LOCAL: Module is in local mode and cannot be used remotely.
    """

    CLOSED = "closed"
    READY = "ready"
    ERROR = "error"
    LOCAL = "local"


class ExposureStatus(Enum):
    """Enumerator for camera status.

    Attributes:
        IDLE: Camera is idle, i.e. ready for work.
        EXPOSING: Camera is currently exposing.
        READOUT: Camera is currently reading out.
        ERROR: Camera is in error state.
    """

    IDLE = "idle"
    EXPOSING = "exposing"
    READOUT = "readout"
    ERROR = "error"


class ImageType(Enum):
    """Enumerator specifying the image type.

    Attributes:
        BIAS: Bias/zero exposure.
        DARK: Dark exposure.
        OBJECT: Object/science exposure
        SKYFLAT: Flat-field taken on sky.
        FOCUS: Exposure from a focus-series.
        ACQUISITION: Exposure from an acquisition
    """

    BIAS = "bias"
    DARK = "dark"
    OBJECT = "object"
    SKYFLAT = "skyflat"
    FOCUS = "focus"
    ACQUISITION = "acquisition"


class ImageFormat(Enum):
    """Enumerator for image formats.

    Attributes:
        INT8: 8 bit integer (i.e. byte).
        INT16: 16 bit integer (i.e. short).
        FLOAT32: 32 bit float.
        FLOAT64: 64 bit float (i.e. double).
        RGB24: RGB format with 8 bit for each colour.
    """

    INT8 = "int8"
    INT16 = "int16"
    FLOAT32 = "float32"
    FLOAT64 = "float64"
    RGB24 = "rgb24"


class MotionStatus(Enum):
    """Enumerator for moving device status.

    Attributes:
        PARKED: The device needs to be initialized or positioned or
            oved (depending upon the device; some devices don't need a formal
            initialization); presumedly, this is the safe "off" state.
        INITIALIZING: The device is transitioning from a PARKED state
            to an active state but is not yet fully operable.
        IDLE: Operating but in no particular state.
        POSITIONED: Operating in a well-defined state, but not moving.
        SLEWING: The device is moving to some targeted state (e.g. to POSITIONED or TRACKING) but has not yet
            arrived at that state.
        TRACKING: The device is moving as commanded.
    """

    ABORTING = "aborting"
    ERROR = "error"
    IDLE = "idle"
    INITIALIZING = "initializing"
    PARKING = "parking"
    PARKED = "parked"
    POSITIONED = "positioned"
    SLEWING = "slewing"
    TRACKING = "tracking"
    UNKNOWN = "unknown"


class WeatherSensors(Enum):
    """Enumerator for sensors of a weather station.

    Attributes:
        TIME: Time of measurement.
        TEMPERATURE: Temperature in °C
        HUMIDITY: Relative humidity in %.
        PRESSURE: Pressure in hPa.
        WINDDIR: Wind direction in degrees azimuth.
        WINDSPEED: Wind speed in km/h.
        RAIN: Raining (1) or not (0).
        SKYTEMP: Relative sky temperature in °C.
        DEWPOINT: Dew point in °C.
        PARTICLES: Number of particles per m³.

    """

    TIME = "time"
    TEMPERATURE = "temp"
    HUMIDITY = "humid"
    PRESSURE = "press"
    WINDDIR = "winddir"
    WINDSPEED = "windspeed"
    RAIN = "rain"
    SKYTEMP = "skytemp"
    DEWPOINT = "dewpoint"
    PARTICLES = "particles"


__all__ = ["ModuleState", "ExposureStatus", "ImageType", "ImageFormat", "MotionStatus", "WeatherSensors"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IReady(Interface, metaclass=ABCMeta):
    """The module can be in a "not ready" state for science and need to be initialized in some way."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def is_ready(self, **kwargs: Any) -> bool:
        """Returns the device is "ready", whatever that means for the specific device.

        Returns:
            Whether device is ready
        """
        ...


__all__ = ["IReady"]



import logging
from typing import Any, Optional
from aiohttp import web

from pyobs.modules import Module
from pyobs.utils.cache import DataCache

log = logging.getLogger(__name__)


class HttpFileCache(Module):
    """A file cache based on a HTTP server."""

    def __init__(self, port: int = 37075, cache_size: int = 25, max_file_size: int = 100, **kwargs: Any):
        """Initializes file cache.

        Args:
            port: Port for HTTP server.
            cache_size: Size of file cache, i.e. number of files to cache.
            max_file_size: Maximum file size in MB.
        """
        Module.__init__(self, **kwargs)

        # store stuff
        self._cache = DataCache(cache_size)
        self._is_listening = False
        self._port = port
        self._cache_size = cache_size
        self._max_file_size = max_file_size * 1024 * 1024

        # define web server
        self._app = web.Application()
        self._app.add_routes([web.get("/{filename}", self.download_handler), web.post("/", self.upload_handler)])
        self._runner = web.AppRunner(self._app)
        self._site: Optional[web.TCPSite] = None

    async def open(self) -> None:
        """Open server"""
        await Module.open(self)

        # start listening
        log.info("Starting HTTP file cache on port %d...", self._port)
        await self._runner.setup()
        self._site = web.TCPSite(self._runner, "0.0.0.0", self._port)
        await self._site.start()
        self._is_listening = True

    async def close(self) -> None:
        """Close server"""
        await Module.close(self)

        # stop server
        await self._runner.cleanup()

    @property
    def opened(self) -> bool:
        """Whether the server is started."""
        return self._is_listening

    async def download_handler(self, request: web.Request) -> web.Response:
        """Handles GET access to /{filename} and returns image.

        Args:
            request: Request to respond to.

        Returns:
            Response containing image.
        """

        # get filename
        filename = request.match_info["filename"]

        # get data
        if filename not in self._cache:
            raise web.HTTPNotFound()
        data = self._cache[filename]

        # send it
        log.info(f"Serving file {filename}.")
        return web.Response(body=data)

    async def upload_handler(self, request: web.Request) -> web.Response:
        """Handles PUSH access to /, stores image and returns filename.

        Args:
            request: Request to respond to.

        Returns:
            Response containing filename.
        """

        # read multipart data
        reader = await request.multipart()
        filename: Optional[str] = None
        data: Optional[bytes] = None
        async for field in reader:
            # we expect a file called 'file'
            if field.name == "file":
                filename = field.filename
                data = await field.read()
                break

        # no filename
        if filename is None:
            raise web.HTTPNotFound()

        # store it
        log.info(f"Storing file {filename}.")
        self._cache[filename] = data

        # return filename
        return web.Response(body=filename)


__all__ = ["HttpFileCache"]



import asyncio
import inspect
import logging
from typing import Union, Type, Any, Callable, Dict, Tuple, List, TypeVar, Optional, cast
from py_expression_eval import Parser
import packaging.version

from pyobs.events import ModuleOpenedEvent, Event
from pyobs.object import Object
from pyobs.interfaces import IModule, IConfig, Interface
from pyobs.utils.enums import ModuleState
from pyobs.utils.types import cast_response_to_simple, cast_bound_arguments_to_real
from pyobs.version import version
from pyobs.utils import exceptions as exc

log = logging.getLogger(__name__)


F = TypeVar("F", bound=Callable[..., Any])


def timeout(func_timeout: Union[str, int, Callable[..., Any], None] = None) -> Callable[[F], F]:
    """Decorates a method with information about timeout for an async HTTP call.

    :param func_timeout:  Integer or string that specifies the timeout.
                          If string, it is parsed using the variables in kwargs.
    """

    def timeout_decorator(func: F) -> F:
        async def _timeout(obj: Any, *args: Any, **kwargs: Any) -> float:
            # define variables as non-local
            nonlocal func_timeout, func

            # init to 0 second
            to = 0.0

            # do we have a timeout?
            if func_timeout is not None:
                # what is it?
                if callable(func_timeout):
                    # this is a method, does it have a timeout on it's own? then use it
                    try:
                        if hasattr(func_timeout, "timeout"):
                            # call timeout method, only works if this has the same parameters
                            to = await getattr(func_timeout, "timeout")(obj, *args, **kwargs)
                        else:
                            # call method directly
                            to = await func_timeout(obj, *args, **kwargs)
                    except:
                        log.exception("Could not call timeout method.")

                elif isinstance(func_timeout, str):
                    # this is a string with a function, so evaluate it
                    try:
                        parser = Parser()
                        to = parser.parse(func_timeout).evaluate(kwargs)
                    except Exception:
                        log.error('Could not find timeout "%s" in list of parameters.', func_timeout)

                else:
                    # it's a number, add timeout directly
                    try:
                        to = float(func_timeout)
                    except ValueError:
                        log.exception("Could not convert timeout to float.")
                        to = 0.0

            # return it
            return to

        # decorate method
        setattr(func, "timeout", _timeout)
        return func

    return timeout_decorator


class Module(Object, IModule, IConfig):
    """Base class for all pyobs modules."""

    __module__ = "pyobs.modules"

    def __init__(self, name: Optional[str] = None, label: Optional[str] = None, **kwargs: Any):
        """
        Args:
            name: Name of module. If None, ID from comm object is used.
            label: Label for module. If None, name is used.
        """
        Object.__init__(self, **kwargs)

        # get list of client interfaces
        self._interfaces: List[Type[Interface]] = []
        self._methods: Dict[str, Tuple[Callable[..., Any], inspect.Signature]] = {}
        self._get_interfaces_and_methods()

        # get configuration caps, i.e. all parameters from c'tor
        self._config_caps = self._get_config_caps()

        # name and label
        self._device_name = name if name is not None else self.comm.name
        self._label = label if label is not None else self._device_name

        # state
        self._state = ModuleState.READY
        self._error_string = ""

    async def open(self) -> None:
        # open comm
        if self.comm is not None:
            # open it and connect module
            self.comm.module = self
            await self.comm.open()

            # react to connecting modules
            await self.comm.register_event(ModuleOpenedEvent, self._on_module_opened)

        """Open module."""
        await Object.open(self)

    async def close(self) -> None:
        """Close module."""
        await Object.close(self)

        # close comm
        if self.comm is not None:
            log.info("Closing connection to server...")
            await self.comm.close()

    async def main(self) -> None:
        """Main loop for application."""
        while True:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                return

    def name(self, **kwargs: Any) -> str:
        """Returns name of module."""
        return "" if self._device_name is None else self._device_name

    async def get_label(self, **kwargs: Any) -> str:
        """Returns label of module."""
        return "" if self._label is None else self._label

    async def get_version(self, **kwargs: Any) -> str:
        """Returns pyobs version of module."""
        return version()

    async def _on_module_opened(self, event: Event, sender: str) -> bool:
        """React to other modules connecting."""
        if sender == self.comm.name or not isinstance(event, ModuleOpenedEvent):
            return False

        # get proxy and version
        proxy = await self.proxy(sender, IModule)
        try:
            module_version = await proxy.get_version()
        except exc.RemoteError:
            return True

        # log it
        log.debug(f"Other module {sender} found, running on pyobs {module_version}.")

        # check version, only compare major and minor, ignore patch level
        v1, v2 = packaging.version.parse(version()), packaging.version.parse(module_version)
        if v1.major != v2.major or v1.minor != v2.minor:
            log.critical(
                f'Found module "{sender}" with different pyobs version {module_version} (≠{version()}), '
                f"please update pyobs."
            )

        # okay
        return True

    @property
    def interfaces(self) -> List[Type[Interface]]:
        """List of implemented interfaces."""
        return self._interfaces

    @property
    def methods(self) -> Dict[str, Tuple[Callable[..., Any], inspect.Signature]]:
        """List of methods."""
        return self._methods

    def _get_interfaces_and_methods(self) -> None:
        """List interfaces and methods of this module."""
        import pyobs.interfaces

        # get interfaces
        self._interfaces = []
        self._methods = {}
        for _, interface in inspect.getmembers(pyobs.interfaces, predicate=inspect.isclass):
            # is module a sub-class of that class that inherits from Interface?
            if isinstance(self, interface) and issubclass(interface, pyobs.interfaces.Interface):
                # we ignore the interface "Interface"
                if interface == pyobs.interfaces.Interface:
                    continue

                # add interface
                self._interfaces += [interface]

                # loop methods of that interface
                for method_name, method in inspect.getmembers(interface, predicate=inspect.isfunction):
                    # get method and signature
                    func = getattr(self, method_name)
                    signature = inspect.signature(func)

                    # fill dict of name->(method, signature)
                    self._methods[method_name] = (func, signature)

    def quit(self) -> None:
        """Quit module."""
        asyncio.get_event_loop().stop()

    async def execute(self, method: str, *args: Any, **kwargs: Any) -> Any:
        """Execute a local method safely with type conversion

        All incoming variables in args and kwargs must be of simple type (i.e. int, float, str, bool, tuple) and will
        be converted to the requested type automatically. All outgoing variables are converted to simple types
        automatically as well.

        Args:
            method: Name of method to execute.
            *args: Parameters for method.
            **kwargs: Parameters for method.

        Returns:
            Response from method call.

        Raises:
            KeyError: If method does not exist.
        """

        # is module in error state?
        if self._state == ModuleState.ERROR:
            # if called method is not from IModule, raise error
            if not hasattr(IModule, method):
                raise exc.ModuleError("Module is in error state, please reset it.")

        # get method and signature (may raise KeyError)
        func, signature = self.methods[method]

        # bind parameters
        ba = signature.bind(*args, **kwargs)
        ba.apply_defaults()

        # cast to types requested by method
        cast_bound_arguments_to_real(ba, signature)

        # get additional args and kwargs and delete from ba
        func_args = []
        func_kwargs = {}
        if "args" in ba.arguments:
            func_args = ba.arguments["args"]
            del ba.arguments["args"]
        if "kwargs" in ba.arguments:
            func_kwargs = ba.arguments["kwargs"]
            del ba.arguments["kwargs"]

        # call method
        response = await func(*func_args, **ba.arguments, **func_kwargs)

        # finished
        return cast_response_to_simple(response)

    def _get_config_caps(self) -> Dict[str, Tuple[bool, bool, bool]]:
        """Returns a dictionary with config caps."""

        # init dict of caps and types
        caps = {}

        # loop super classes
        for cls in inspect.getmro(self.__class__):
            # ignore Object and Module
            if cls in [Object, Module]:
                continue

            # get signature
            sig = inspect.signature(getattr(cls, "__init__"))
            for name in sig.parameters:
                # ignore self, args, kwargs
                if name in ["self", "args", "kwargs"]:
                    continue

                # check for getter and setter
                caps[name] = (
                    hasattr(self, "_get_config_" + name),
                    hasattr(self, "_set_config_" + name),
                    hasattr(self, "_get_config_options_" + name),
                )

        # finished
        return caps

    async def get_config_caps(self, **kwargs: Any) -> Dict[str, Tuple[bool, bool, bool]]:
        """Returns dict of all config capabilities. First value is whether it has a getter, second is for the setter,
        third is for a list of possible options..

        Returns:
            Dict with config caps
        """
        return self._config_caps

    async def get_config_value(self, name: str, **kwargs: Any) -> Any:
        """Returns current value of config item with given name.

        Args:
            name: Name of config item.

        Returns:
            Current value.

        Raises:
            ValueError: If config item of given name does not exist.
        """

        # valid parameter?
        if name not in self._config_caps:
            raise ValueError("Invalid parameter %s" % name)
        if not self._config_caps[name][0]:
            raise ValueError("Parameter %s is not remotely accessible.")

        # get getter method and call it
        getter = getattr(self, "_get_config_" + name)
        return getter()

    async def get_config_value_options(self, name: str, **kwargs: Any) -> List[str]:
        """Returns possible values for config item with given name.

        Args:
            name: Name of config item.

        Returns:
            Possible values.

        Raises:
            ValueError: If config item of given name does not exist.
        """

        # valid parameter?
        if name not in self._config_caps:
            raise ValueError("Invalid parameter %s" % name)
        if not self._config_caps[name][2]:
            raise ValueError("Parameter %s has no list of possible values.")

        # get getter method and call it
        options = getattr(self, "_get_config_options_" + name)
        return cast(List[str], options())

    async def set_config_value(self, name: str, value: Any, **kwargs: Any) -> None:
        """Sets value of config item with given name.

        Args:
            name: Name of config item.
            value: New value.

        Raises:
            ValueError: If config item of given name does not exist or value is invalid.
        """

        # valid parameter?
        if name not in self._config_caps:
            raise ValueError("Invalid parameter %s" % name)
        if not self._config_caps[name][1]:
            raise ValueError("Parameter %s is not remotely settable.")

        # get setter and call it
        setter = getattr(self, "_set_config_" + name)
        setter(value)

    async def set_state(self, state: ModuleState, error_string: Optional[str] = None) -> None:
        """Set state of module.

        Args:
            state: New state to set.
            error_string: If given, set error string.
        """

        # log?
        if state != self._state:
            log.info("Set module state to %s.", state)

        # set it
        self._state = state
        if error_string is not None:
            self.set_error_string(error_string)

    async def get_state(self, **kwargs: Any) -> ModuleState:
        """Returns current state of module."""
        return self._state

    async def reset_error(self, **kwargs: Any) -> bool:
        """Reset error of module, if any. Should be overwritten by derived class to handle error resolution."""
        self._state = ModuleState.READY
        self.set_error_string()
        return True

    async def _default_remote_error_callback(self, exception: exc.PyObsError) -> None:
        """Called on severe errors.

        Args:
            exception: Exception that caused severe error.
        """

        # set error string
        if isinstance(exception, exc.RemoteError):
            error = f"Servere error in {exception.module} module: {exception}"
        else:
            error = f"Severe error: {exception}"
        self.set_error_string(error)

        # log it and set state
        log.critical(error)
        await self.set_state(ModuleState.ERROR)

    def set_error_string(self, error: str = "") -> None:
        """Set error string."""
        self._error_string = error

    async def get_error_string(self, **kwargs: Any) -> str:
        """Returns description of error, if any."""
        return self._error_string


class MultiModule(Module):
    """Wrapper for running multiple modules in a single process."""

    __module__ = "pyobs.modules"

    def __init__(
        self,
        modules: Dict[str, Union[Module, Dict[str, Any]]],
        shared: Optional[Dict[str, Union[object, Dict[str, Any]]]] = None,
        **kwargs: Any,
    ):
        """
        Args:
            modules: Dictionary with modules.
            shared: Shared objects between modules.
        """
        Module.__init__(self, name="multi", **kwargs)

        # create shared objects
        self._shared: Dict[str, Module] = {}
        if shared:
            for name, obj in shared.items():
                # if obj is an object definition, create it, otherwise just set it
                self._shared[name] = self.add_child_object(obj)

        # create modules
        self._modules = {}
        for name, mod in modules.items():
            # what is it?
            if isinstance(mod, Module):
                # it's a module already, store it
                self._modules[name] = mod
            elif isinstance(mod, dict):
                # dictionary, create it
                self._modules[name] = self.add_child_object(mod, **self._shared, copy_comm=False)

    @property
    def modules(self) -> Dict[str, Module]:
        return self._modules

    def __contains__(self, name: str) -> bool:
        """Checks, whether this multi-module contains a module of given name."""
        return name in self._modules

    def __getitem__(self, name: str) -> Module:
        """Returns module of given name."""
        return self._modules[name]


__all__ = ["Module", "MultiModule", "timeout"]



import itertools
import re
from typing import Dict, Optional, Union, List, Tuple, Any
from astroplan import Observer
from astropy.time import TimeDelta
from py_expression_eval import Parser, Expression
import logging
import astropy.units as u

from pyobs.utils.time import Time


log = logging.getLogger(__name__)


class ExpTimeEval:
    """Exposure time evaluator for skyflats."""

    def __init__(self, observer: Observer, functions: Union[str, Dict[str, Union[str, Dict[str, str]]]]):
        """Initializes a new evaluator.

        Args:
            observer: Observer to use.
            functions: Dict of functions for the different filters/binnings.
                Three possible formats:
                1. Just a string with a function, e.g. 'exp(-0.9*(h+3.9))', completely ignoring binning and filter.
                2. Dictionary on filter or binning like
                   {'1x1': 'exp(-0.9*(h+3.9))'}
                   or
                   {'clear': 'exp(-0.9*(h+3.9))'}
                   If a binning is given, filters are ignored, and vice versa. Binnings need to be given as NxN.
                3. Nested dictionary with binning and filter like
                   {'1x1': {'clear': 'exp(-0.9*(h+3.9))'}}
                   In this structure, binning must be the first layer, followed by filter.
        """

        # init
        self._observer = observer
        self._time: Optional[Time] = None
        self._m: Optional[float] = None
        self._b: Optional[float] = None

        # get parser and init functions dict, which is a tuple holding x/y binning and filter
        p = Parser()
        self._functions: Dict[Tuple[Optional[Tuple[int, int]], Optional[str]], Expression] = {}

        # so, what format is the functions dict?
        if isinstance(functions, str):
            # single function
            self._functions[None, None] = p.parse(functions)
            log.info("Found a single flatfield function for all binnings and filters.")

        else:
            # check, whether keys are binnings or filters
            is_binning = [re.match("[0-9]+x[0-9]+", k) is not None for k in functions.keys()]
            if any(is_binning) and not all(is_binning):
                raise ValueError("Inconsistent configuration: first layer is neither all binnings nor all filters. ")

            # if all entries in is_binning are True, first layer is binnings
            if all(is_binning):
                # 1st level is binnings, is next level strings or another dict?
                is_str = [isinstance(f, str) for f in functions.values()]
                if any(is_str) and not all(is_str):
                    raise ValueError("Inconsistent configuration: second layer is neither all str nor all dicts.")

                # filters or not?
                if all(is_str):
                    # all strings, so we don't have filters
                    for b, func in functions.items():
                        self._functions[self._bin(b), None] = p.parse(func)

                else:
                    # need to go a level deeper
                    for b, tmp in functions.items():
                        if isinstance(tmp, dict):
                            for f, func in tmp.items():
                                self._functions[self._bin(b), f] = p.parse(func)
                        else:
                            raise ValueError("Inconsistent configuration.")

            else:
                # 1st level is filters, second level must be strings!
                is_str = [isinstance(f, str) for f in functions.values()]
                if not all(is_str):
                    raise ValueError("Inconsistent configuration: second level must be functions.")

                # parse
                self._functions = {(None, f): p.parse(func) for f, func in functions.items()}

    @staticmethod
    def _bin(binning: str) -> Tuple[int, int]:
        """Split binning"""
        s = binning.split("x")
        return int(s[0]), int(s[1])

    def _keys(self, i: int) -> List[Any]:
        keys = list(set([k[i] for k in self._functions.keys()]))
        if None in keys:
            keys.remove(None)
        return sorted(keys)

    @property
    def binnings(self) -> List[Tuple[int, int]]:
        """Return list of binnings."""
        return self._keys(0)

    @property
    def filters(self) -> List[str]:
        """Return list of filters."""
        return self._keys(1)

    def __call__(
        self, solalt: float, binning: Optional[Tuple[int, int]] = None, filter_name: Optional[str] = None
    ) -> float:
        """Estimate exposure time for given filter

        Args:
            solalt: Solar altitude.
            binning: Used binning.
            filter_name: Name of filter.

        Returns:
            Estimated exposure time.
        """

        # got binning?
        got_binnings = len(self.binnings) > 0

        # if we got no filters, ignore filter_name, if given
        if len(self.filters) == 0:
            filter_name = None

        # get function and evaluate it
        exptime = float(self._functions[binning if got_binnings else None, filter_name].evaluate({"h": solalt}))

        # need to scale with exp time?
        return exptime / (binning[0] * binning[1]) if not got_binnings and binning is not None else exptime

    def init(self, time: Time) -> None:
        """Initialize object with the given time.

        Args:
            time: Start time for all further calculations.
        """

        # store time
        self._time = time

        # get sun now and in 10 minutes
        sun_now = self._observer.sun_altaz(time)
        sun_10min = self._observer.sun_altaz(time + TimeDelta(10 * u.minute))

        # get m, b for calculating sun_alt=m*time+b
        self._b = sun_now.alt.degree
        self._m = (sun_10min.alt.degree - self._b) / (10.0 * 60.0)

    def exp_time(
        self, time_offset: float, binning: Optional[Tuple[int, int]] = None, filter_name: Optional[str] = None
    ) -> float:
        """Estimates exposure time for a given filter and binning at a given time offset from the start time (see init).

        Args:
            time_offset: Offset in seconds from start time (see init)
            binning: Used binning.
            filter_name: Name of filter

        Returns:
            Estimated exposure time
        """
        if self._m is None or self._b is None:
            raise ValueError("m or b not set.")
        return self(self._m * time_offset + self._b, binning=binning, filter_name=filter_name)

    def duration(
        self,
        count: int,
        start_time: float = 0,
        readout: float = 0,
        binning: Optional[Tuple[int, int]] = None,
        filter_name: Optional[str] = None,
    ) -> float:
        """Estimates the duration for a given amount of flats in the given filter and binning, starting at the given
        start time.

        Args
            count: Number of flats to take.:
            start_time: Time in seconds to start after the time set in init()
            readout: Time in seconds for readout per flat
            binning: Used binning.
            filter_name: Name of filter

        Returns:
            Estimated duration in seconds
        """

        # loop through images and add estimated exposure times at their respective start times
        elapsed = start_time
        for i in range(count):
            elapsed += self.exp_time(elapsed, binning=binning, filter_name=filter_name) + readout

        # we started at start_time, so subtract it again
        return elapsed - start_time



import logging
from typing import Any, List, Dict

from pyobs.events import Event
from pyobs.modules import Module
from pyobs.interfaces import IAutonomous
from pyobs.object import get_class_from_string

log = logging.getLogger(__name__)


class Trigger(Module, IAutonomous):
    """A module that can call another module's methods when a specific event occurs."""

    __module__ = "pyobs.modules.utils"

    def __init__(self, triggers: List[Dict[str, Any]], **kwargs: Any):
        """Initialize a new trigger module.

        Args:
            triggers: List of dictionaries defining the trigger. Must contain fields for event, module and method,
                      may contain a sender.

        """
        Module.__init__(self, **kwargs)

        # store
        self._running = False

        # store triggers and convert event strings to actual classes
        self._triggers = triggers
        for trigger in self._triggers:
            # get class and store it
            kls = get_class_from_string(trigger["event"])
            trigger["event"] = kls

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # get a list of all events
        events = list(set([t["event"] for t in self._triggers]))

        # start
        self._running = True

        # register them
        for event in events:
            await self.comm.register_event(event, self._handle_event)

    async def start(self, **kwargs: Any) -> None:
        """Starts a service."""
        self._running = True

    async def stop(self, **kwargs: Any) -> None:
        """Stops a service."""
        self._running = False

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether a service is running."""
        return self._running

    async def _handle_event(self, event: Event, sender: str) -> bool:
        """Handle an incoming event.

        Args:
            event: The received event
            sender: Name of sender
        """

        # not running?
        if not self._running:
            return False

        # loop all triggers
        for trigger in self._triggers:
            # does it handle the event?
            if trigger["event"] == event.__class__:
                log.info(
                    "Received a %s event and calling %s.%s now.", str(type(event)), trigger["module"], trigger["method"]
                )

                # get proxy
                try:
                    proxy = await self.comm.proxy(trigger["module"])

                    # call it
                    await proxy.execute(trigger["method"])

                except ValueError:
                    log.exception("Could not execute command on proxy %s.", trigger["module"])
                    continue

                except Exception as e:
                    log.error("Error on calling %s.%s: %s", trigger["module"], trigger["method"], e)
                    continue

        return True


__all__ = ["Trigger"]



import logging
from typing import Any

from .publisher import Publisher


log = logging.getLogger(__name__)


class LogPublisher(Publisher):
    def __init__(self, level: str = "info", **kwargs: Any):
        """Initialize new log publisher.

        Args:
            level: Level to log on.
        """
        Publisher.__init__(self, **kwargs)

        # set and check level
        if not hasattr(log, level):
            raise ValueError("Unknown log level.")
        self._log_function = getattr(log, level)

    async def __call__(self, **kwargs: Any) -> None:
        """Publish the given results.

        Args:
            **kwargs: Results to publish.
        """

        # build string
        s = ", ".join(["{0}={1}".format(k, v) for k, v in kwargs.items()])

        # log it
        self._log_function(s)


__all__ = ["LogPublisher"]



from __future__ import annotations

import asyncio
from enum import Enum
import logging
from typing import Dict, Union, Callable, Optional, Tuple, Any, cast, Coroutine
import astropy.units as u
from astropy.time import TimeDelta
import numpy as np

from pyobs.object import Object
from pyobs.utils.enums import ImageType
from pyobs.utils.fits import fitssec
from pyobs.utils.parallel import Future, event_wait
from pyobs.utils.time import Time
from .exptimeeval import ExpTimeEval
from .pointing import SkyFlatsBasePointing
from pyobs.interfaces import ITelescope, ICamera, IFilters, IExposureTime, IBinning, IWindow, IImageType
from pyobs.images import Image

log = logging.getLogger(__name__)


class FlatFielder(Object):
    """Automatized flat-fielding."""

    __module__ = "pyobs.utils.skyflats"

    class Twilight(Enum):
        DUSK = "dusk"
        DAWN = "dawn"

    class State(Enum):
        INIT = "init"
        WAITING = "waiting"
        TESTING = "testing"
        RUNNING = "running"
        FINISHED = "finished"

    def __init__(
        self,
        functions: Union[str, Dict[str, Union[str, Dict[str, str]]]],
        target_count: float = 30000,
        min_exptime: float = 0.5,
        max_exptime: float = 5,
        test_frame: Optional[Tuple[float, float, float, float]] = None,
        counts_frame: Optional[Tuple[float, float, float, float]] = None,
        allowed_offset_frac: float = 0.2,
        min_counts: int = 100,
        pointing: Optional[Union[Dict[str, Any], SkyFlatsBasePointing]] = None,
        callback: Optional[Callable[..., Coroutine[Any, Any, None]]] = None,
        **kwargs: Any,
    ):
        """Initialize a new flat fielder.

        Args:
            functions: Function f(h) for each filter to describe ideal exposure time as a function of solar
                elevation h, i.e. something like exp(-0.9*(h+3.9)). See ExpTimeEval for details.
            target_count: Count rate to aim for.
            min_exptime: Minimum exposure time.
            max_exptime: Maximum exposure time.
            test_frame: Tupel (left, top, width, height) in percent that describe the frame for on-sky testing.
            counts_frame: Tupel (left, top, width, height) in percent that describe the frame for calculating mean
                count rate.
            allowed_offset_frac: Offset from target_count (given in fraction of it) that's still allowed for good
                flat-field
            min_counts: Minimum counts in frames.
            observer: Observer to use.
            vfs: VFS to use.
            callback: Callback function for statistics.
        """
        Object.__init__(self, **kwargs)

        # store stuff
        self._target_count = target_count
        self._min_exptime = min_exptime
        self._max_exptime = max_exptime
        self._test_frame = (45, 45, 10, 10) if test_frame is None else test_frame
        self._counts_frame = (25, 25, 75, 75) if counts_frame is None else counts_frame
        self._allowed_offset_frac = allowed_offset_frac
        self._min_counts = min_counts
        self._callback = callback

        # parse function
        self._eval = ExpTimeEval(self.observer, functions)

        # abort event
        self._abort = asyncio.Event()

        # pointing
        self._pointing: Optional[SkyFlatsBasePointing] = self.get_safe_object(pointing, SkyFlatsBasePointing)

        # state machine
        self._state = FlatFielder.State.INIT

        # current exposure time
        self._exptime = 0.0

        # median of last image
        self._median = 0.0

        # exposures to do
        self._exposures_total = 0
        self._exposures_done = 0
        self._exptime_done = 0.0

        # bias level
        self._bias_level = 0.0

        # which twilight are we in? just init something
        self._twilight = FlatFielder.Twilight.DAWN

        # current request
        self._cur_filter: Optional[str] = None
        self._cur_binning: Tuple[int, int] = (1, 1)

    async def __call__(
        self,
        telescope: ITelescope,
        camera: ICamera,
        filters: Optional[IFilters] = None,
        filter_name: Optional[str] = None,
        count: int = 20,
        binning: Tuple[int, int] = (1, 1),
    ) -> State:
        """Calls next step in state machine.

        Args:
            telescope: Telescope to use.
            camera: Camera to use.
            filters: Filter wheel to use.
            filter_name: Name of filter to do flat fields in.
            count: Number of flat fields to take.
            binning: Binning for flat fields.

        Returns:
            Current state of flat fielder.
        """

        # camera must support exposure times
        if not isinstance(camera, IExposureTime):
            raise ValueError("Camera must support exposure times.")

        # store
        self._cur_filter = filter_name
        self._cur_binning = binning
        self._exposures_total = count

        # which state are we in?
        if self._state == FlatFielder.State.INIT:
            # init task
            await self._init_system(telescope, camera, filters)
        elif self._state == FlatFielder.State.WAITING:
            # wait until exposure time reaches good time
            await self._wait()
        elif self._state == FlatFielder.State.TESTING:
            # do actual tests on sky for exposure time
            await self._testing(cast(ICamera, camera))
        elif self._state == FlatFielder.State.RUNNING:
            # take flat fields
            await self._flat_field(telescope, cast(ICamera, camera))

        # return current state
        return self._state

    async def reset(self) -> None:
        """Reset flat fielder"""
        self._state = FlatFielder.State.INIT
        self._exposures_done = 0
        self._exptime_done = 0
        if self._pointing is not None:
            await self._pointing.reset()

    @property
    def has_filters(self) -> bool:
        """Returns True, if functions are based on filters."""
        return len(self._eval.filters) > 0

    @property
    def image_count(self) -> int:
        return self._exposures_done

    @property
    def total_exptime(self) -> float:
        return self._exptime_done

    def _initial_check(self) -> bool:
        """Do a quick initial check.

        Returns:
            False, if flat-field time for this filter is over, True otherwise.
        """

        # get solar elevation and evaluate function
        sun_alt, self._exptime = self._eval_function(Time.now())
        log.info(
            "Calculated optimal exposure time of %.2fs in %dx%d at solar elevation of %.2f°.",
            self._exptime,
            self._cur_binning[0],
            self._cur_binning[1],
            sun_alt,
        )

        # then evaluate exposure time within a larger range
        state = self._eval_exptime(self._min_exptime * 0.5, self._max_exptime * 2.0)
        if state > 0:
            log.info("Missed flat-fielding time, finishing task...")
            self._state = FlatFielder.State.FINISHED
            return False
        else:
            log.info("Flat-field time is still coming, keep going...")
            return True

    async def _init_system(
        self, telescope: ITelescope, camera: Union[ICamera, IExposureTime], filters: Optional[IFilters] = None
    ) -> None:
        """Initialize whole system."""

        # which twilight are we in?
        if self.observer is None:
            raise ValueError("No observer given.")
        sun = self.observer.sun_altaz(Time.now())
        sun_10min = self.observer.sun_altaz(Time.now() + TimeDelta(10 * u.minute))
        self._twilight = (
            FlatFielder.Twilight.DUSK if sun_10min.alt.degree < sun.alt.degree else FlatFielder.Twilight.DAWN
        )
        log.info("We are currently in %s twilight.", self._twilight.value)

        # do initial check
        if not self._initial_check():
            return

        # set binning
        if isinstance(camera, IBinning):
            log.info("Setting binning to %dx%d...", self._cur_binning[0], self._cur_binning[1])
            await camera.set_binning(*self._cur_binning)

        # get bias level
        self._bias_level = await self._get_bias(camera)

        # move telescope
        future_track = self._pointing(telescope) if self._pointing is not None else None

        # get filter from first step and set it
        if filters is not None and self._cur_filter is not None:
            log.info("Setting filter to %s...", self._cur_filter)
            future_filter = await filters.set_filter(self._cur_filter)
        else:
            future_filter = Future(empty=True)

        # wait for both
        await Future.wait_all([future_track, future_filter])
        log.info("Finished initializing system.")

        # change stats
        log.info("Waiting for flat-field time...")
        self._state = FlatFielder.State.WAITING

    async def _get_bias(self, camera: ICamera) -> float:
        """Take bias image to determine bias level.

        Returns:
            Median bias level.
        """
        log.info("Taking BIAS image to determine median level...")

        # set full frame
        cam = camera
        if isinstance(camera, IWindow):
            full_frame = await camera.get_full_frame()
            await camera.set_window(*full_frame)

        # take image
        if isinstance(camera, IExposureTime):
            await camera.set_exposure_time(0.0)
        if isinstance(camera, IImageType):
            await camera.set_image_type(ImageType.BIAS)
        filename = await cam.grab_image(broadcast=False)

        # download image
        bias = await self.vfs.read_image(filename)
        avg = float(np.median(bias.data))

        # return it
        log.info("Found median BIAS level of %.2f...", avg)
        return avg

    async def _wait(self) -> None:
        """Wait for flat-field time."""

        # get solar elevation and evaluate function
        sun_alt, self._exptime = self._eval_function(Time.now())
        log.info(
            "Calculated optimal exposure time of %.2fs in %dx%d at solar elevation of %.2f°.",
            self._exptime,
            self._cur_binning[0],
            self._cur_binning[1],
            sun_alt,
        )

        # then evaluate exposure time within a larger range
        state = self._eval_exptime(self._min_exptime * 0.5, self._max_exptime * 2.0)
        if state < 0:
            log.info("Sleeping a little...")
            await event_wait(self._abort, 10)
        elif state == 0:
            log.info("Starting to take test flat-fields...")
            self._state = FlatFielder.State.TESTING
        else:
            log.info("Missed flat-fielding time, finish task...")
            self._state = FlatFielder.State.FINISHED

    def _eval_function(self, time: Time) -> Tuple[float, float]:
        """Evaluate function for given filter at given time.

        Args:
            time: Time to evaluate function at.

        Returns:
            Estimated exposure time.
        """

        # get solar elevation and evaluate exptime
        if self.observer is None:
            raise ValueError("No observer given.")
        sun = self.observer.sun_altaz(time)
        exptime = self._eval(sun.alt.degree, binning=self._cur_binning, filter_name=self._cur_filter)

        # return solar altitude and exposure time
        return float(sun.alt.degree), exptime

    def _eval_exptime(self, min_exptime: Optional[float] = None, max_exptime: Optional[float] = None) -> int:
        """Evaluates current exposure time. Sets new state or waits of necessary.

        Returns:
            -1, if we have to wait, 0 during flat-field time and 1 if it has passed.
        """

        # no min/max given?
        if min_exptime is None:
            min_exptime = self._min_exptime
        if max_exptime is None:
            max_exptime = self._max_exptime

        # need to wait, change status or are we finished?
        if (self._twilight == FlatFielder.Twilight.DUSK and self._exptime > max_exptime) or (
            self._twilight == FlatFielder.Twilight.DAWN and self._exptime < min_exptime
        ):
            # in DUSK, if exptime is greater than max exptime, we're past flatfielding time
            # in DAWN, if exptime is less than min exptime, we're past flatfielding time
            return 1
        elif (self._twilight == FlatFielder.Twilight.DUSK and self._exptime < min_exptime) or (
            self._twilight == FlatFielder.Twilight.DAWN and self._exptime > max_exptime
        ):
            # in DUSK, if exptime is less than max exptime, we still need to wait
            # in DAWN, if exptime is greater than min exptime, we still need to wait
            return -1
        else:
            # otherwise it seems that we're in the middle of flat-fielding time
            return 0

    async def _testing(self, camera: ICamera) -> None:
        """Take flat-fields but don't store them."""

        # set window
        await self._set_window(camera, testing=True)

        # do exposures, do not broadcast while testing
        log.info("Exposing test flat field for %.2fs...", self._exptime)
        cam = camera
        if isinstance(camera, IExposureTime):
            await camera.set_exposure_time(float(self._exptime))
        if isinstance(camera, IImageType):
            await camera.set_image_type(ImageType.SKYFLAT)
        filename = await cam.grab_image(broadcast=False)

        # analyse image
        await self._analyse_image(filename)

        # then evaluate exposure time
        state = self._eval_exptime()
        if state < 0:
            log.info("Sleeping a little...")
            await event_wait(self._abort, 10)
        elif state == 0:
            log.info("Starting to store flat-fields...")
            self._state = FlatFielder.State.RUNNING
        else:
            log.info("Missed flat-fielding time, finish task...")
            self._state = FlatFielder.State.FINISHED

    async def _set_window(self, camera: Union[ICamera, IExposureTime], testing: bool) -> None:
        """Set camera window.

        Args:
            testing: Whether we're in testing mode or not.
        """
        if isinstance(camera, IWindow):
            # get full frame
            left, top, width, height = await camera.get_full_frame()

            # if testing, take test frame, otherwise use full frame
            if testing:
                left, top, width, height = (
                    int(left + self._test_frame[0] / 100 * width),
                    int(top + self._test_frame[1] / 100 * width),
                    int(self._test_frame[2] / 100 * width),
                    int(self._test_frame[3] / 100 * height),
                )

            # set it
            log.info("Setting camera window to %dx%d at %d,%d...", width, height, left, top)
            await camera.set_window(left, top, width, height)

    async def _analyse_image(self, filename: str) -> bool:
        """Analyze image and return whether it's okay.

        Args:
            filename: Filename of image.

        Returns:
            Whether flat-field is okay.
        """

        # download image
        flat_field = await self.vfs.read_image(filename)
        if flat_field is None:
            return False

        # get median from image
        self._median = self._get_image_median(flat_field, self._counts_frame)
        log.info("Got a flat field with median counts of %.2f.", self._median)

        # if count rate is too low, don't use this image to calculate new exposure time
        if self._median < self._min_counts:
            log.warning("Median counts (%d) too low, retrying last image with same exposure time...", self._median)
            return False

        else:
            # calculate deviation from target counts
            frac = abs(1.0 - self._median / self._target_count)

            # calculate new exposure time
            self._calc_new_exptime()

            # log and return
            if frac > self._target_count:
                log.warning("Deviation from target count (%.1f%%) is larger than allowed, retrying last image...", frac)
                return False
            else:
                log.info("Calculated new exposure time to be %.2fs.", self._exptime)
                return True

    @staticmethod
    def _get_image_median(image: Image, frame: Optional[Tuple[float, float, float, float]] = None) -> float:
        """Returns median of image after trimming it to TRIMSEC and to given frame.

        Args:
            image: Image to calculate median for.
            frame: Frame coordinates as (width, top, width, height) in percent of full frame.

        Returns:
            Median of image.
        """

        # trim image to TRIMSEC
        data = fitssec(image, "TRIMSEC")

        # cut to frame
        if frame is not None:
            width, height = data.shape
            data = data[
                int(frame[0] / 100 * width) : int((frame[0] + frame[2]) / 100 * width),
                int(frame[1] / 100 * height) : int((frame[1] + frame[3]) / 100 * height),
            ]

        # return median
        return float(np.median(data))

    def _calc_new_exptime(self) -> None:
        """Calculate new exposure time."""
        # calculate factor for new exposure time
        factor = (self._target_count - self._bias_level) / (self._median - self._bias_level)

        # limit factor to 0.1-10
        factor = min(10.0, max(0.1, factor))

        # calculate next exposure time
        self._exptime *= factor

    async def _flat_field(self, telescope: ITelescope, camera: ICamera) -> None:
        """Take flat-fields."""

        # set window
        await self._set_window(camera, testing=False)

        # move telescope
        if self._pointing is not None:
            await self._pointing(telescope)

        # do exposures, do not broadcast while testing
        now = Time.now()
        log.info(
            "Exposing flat field %d/%d for %.2fs...", self._exposures_done + 1, self._exposures_total, self._exptime
        )
        if isinstance(camera, IExposureTime):
            await camera.set_exposure_time(float(self._exptime))
        if isinstance(camera, IImageType):
            await camera.set_image_type(ImageType.SKYFLAT)
        filename = await camera.grab_image()

        # analyse image
        if await self._analyse_image(filename):
            # increase count and quite here, if finished
            self._exptime_done += self._exptime
            self._exposures_done += 1
            if self._exposures_done >= self._exposures_total:
                log.info("Finished all requested flat-fields..")
                self._state = FlatFielder.State.FINISHED
                return

            # call callback
            if self._callback is not None:
                if self.observer is None:
                    raise ValueError("No observer given.")
                sun = self.observer.sun_altaz(now)
                await self._callback(
                    datetime=now.isot,
                    solalt=sun.alt.degree,
                    exptime=self._exptime,
                    counts=self._target_count,
                    filter_name=self._cur_filter,
                    binning=self._cur_binning,
                )

        # then evaluate exposure time
        state = self._eval_exptime()
        if state < 0:
            log.info("Going back to testing...")
            self._state = FlatFielder.State.TESTING
        elif state == 0:
            pass
        else:
            log.info("Missed flat-fielding time, finish task...")
            self._state = FlatFielder.State.FINISHED

    async def abort(self) -> None:
        """Abort current actions."""
        self._abort.set()


__all__ = ["FlatFielder"]



from __future__ import annotations
import inspect
import types
from typing import TYPE_CHECKING, Any, Type, List, Dict

from pyobs.interfaces import Interface
from pyobs.utils.types import cast_bound_arguments_to_simple
from pyobs.utils import exceptions as exc

if TYPE_CHECKING:
    from pyobs.comm import Comm


class Proxy:
    """A proxy for remote pyobs modules."""

    __module__ = "pyobs.comm"

    def __init__(self, comm: Comm, client: str, interfaces: List[Type[Interface]]):
        """Creates a new proxy.

        Args:
            comm: Comm object to use for connection.
            client: Name of client to connect to.
            interfaces: List of interfaces supported by client.
        """

        # set client and interfaces
        self._comm: Comm = comm
        self._client = client
        self._interfaces = interfaces

        # remove interfaces that are implemented by others
        to_delete = []
        for i1 in interfaces:
            for i2 in interfaces:
                if i1 != i2 and issubclass(i1, i2):
                    # i1 implements i2, so remove i2
                    to_delete.append(i2)
        interfaces = [i for i in interfaces if i not in to_delete]

        # add interfaces as base classes
        cls = self.__class__
        self.__class__ = cls.__class__("Proxy", tuple([cls] + interfaces), {})  # type: ignore

        # create methods
        self._methods = self._create_methods()

    @property
    def name(self) -> str:
        """Name of the client."""
        return self._client

    @property
    def method_names(self) -> List[str]:
        """List of method names."""
        return list(sorted(self._methods.keys()))

    @property
    def interfaces(self) -> List[Type[Interface]]:
        """List of interfaces."""
        return self._interfaces

    def signature(self, method: str) -> inspect.Signature:
        """Returns the signature of a given method.

        Args:
            method: Name of the method.

        Returns:
            Signature of the given method.
        """
        return inspect.signature(self._methods[method][0])

    def interface_method(self, method: str) -> Any:
        """Returns the method of the given name from the interface and not from the object itself.

        Args:
            method: Name of method.

        Returns:
            The interface method.
        """
        return self._methods[method][0]

    async def execute(self, method: str, *args: Any, **kwargs: Any) -> Any:
        """Execute a method on the remote client.

        Args:
            method: Name of method to call.
            *args: Parameters for  method call.
            **kwargs: Parameters for method call.

        Returns:
            Result of method call.
        """

        # add 'self' to args
        args = tuple([self] + list(args))

        # get method signature and bind it
        try:
            signature = inspect.signature(self._methods[method][0])
        except KeyError:
            raise exc.RemoteError(self.name, f"Module {self.name} has no method {method}.")
        ba = signature.bind(*args, **kwargs)
        ba.apply_defaults()

        # cast to simple types
        cast_bound_arguments_to_simple(ba)

        # do request and return future
        return await self._comm.execute(self._client, method, signature, *ba.args[1:])

    def _create_methods(self) -> Dict[str, Any]:
        """Create local methods for the remote client."""

        # loop all interfaces and get methods
        methods = {}
        for interface in self._interfaces:
            # loop all methods:
            for func_name, func in inspect.getmembers(interface, predicate=inspect.isfunction):
                # set method
                my_func = types.MethodType(self._remote_function_wrapper(func_name), self)
                setattr(self, func_name, my_func)

                # store func
                methods[func_name] = (func, getattr(self, func_name))

        # return methods
        return methods

    def _remote_function_wrapper(self, method: str) -> Any:
        """Function wrapper for remote calls.

        Args:
            method: Name of method to wrap.

        Returns:
            Wrapper.
        """

        async def inner(this: Proxy, *args: Any, **kwargs: Any) -> Any:
            return await this.execute(method, *args, **kwargs)

        return inner


__all__ = ["Proxy"]



from typing import Tuple

from astropy.coordinates import SkyCoord, BaseCoordinateFrame, Angle


class SkyOffsets:
    def __init__(self, coord0: SkyCoord, coord1: SkyCoord):
        self.coord0 = coord0
        self.coord1 = coord1

    def separation(self, frame: BaseCoordinateFrame = None) -> Angle:
        """Returns separatation between both coordinates, either in their own or a given frame.

        Args:
            frame: Coordinate frame to use, or None to use coordinates' own frames.

        Returns:
            Angle between coordinates.
        """

        # convert and return separation
        coord0, coord1 = self._to_frame(frame)
        return coord0.separation(coord1)

    def spherical_offsets(self, frame: BaseCoordinateFrame = None) -> Tuple[Angle, Angle]:
        """Calculates spherical offset from first coordinate to second.

        Args:
            frame: Coordinate frame to use, or None to use coordinates' own frames.

        Returns:
            Two angles for offset in lat and lon.
        """

        # convert and return offset
        coord0, coord1 = self._to_frame(frame)
        off0, off1 = coord0.spherical_offsets_to(coord1.frame)
        return off0, off1

    def _to_frame(self, frame: BaseCoordinateFrame = None) -> Tuple[SkyCoord, SkyCoord]:
        """

        Args:
            frame: Coordinate frame to use, or None to use coordinates' own frames.

        Returns:
            Both coordinates converted to given frame (or originals, if no frame is given).
        """
        if frame is None:
            return self.coord0, self.coord1
        else:
            return self.coord0.transform_to(frame), self.coord1.transform_to(frame)


__all__ = ["SkyOffsets"]



import datetime
from typing import List, Optional, Tuple

import numpy as np


class RollingTimeAverage(object):
    def __init__(self, interval: float):
        self._interval = interval
        self._values: List[Tuple[datetime.datetime, float]] = []
        self._start_time = datetime.datetime.utcnow()

    def clear(self) -> None:
        self._values = []
        self._start_time = datetime.datetime.utcnow()

    def add(self, value: float) -> None:
        # add value
        now = datetime.datetime.utcnow()
        self._values.append((now, value))

        # clean up
        self._values = [(time, value) for time, value in self._values if (now - time).total_seconds() < self._interval]

    def average(self, min_interval: Optional[float] = None) -> Optional[float]:
        # got values?
        if len(self._values) == 0:
            return None

        # get time
        now = datetime.datetime.utcnow()

        # go no values older than now-interval?
        if min_interval:
            values = [value for time, value in self._values if (now - time).total_seconds() > min_interval]
            if len(values) == 0:
                return None

        # get values and average
        values = [value for time, value in self._values if (now - time).total_seconds() < self._interval]
        return float(np.mean(values))


__all__ = ["RollingTimeAverage"]



from typing import Optional, Any
from typing_extensions import TypedDict

from .event import Event


DataType = TypedDict("DataType", {"focus": float, "error": Optional[float], "filter_name": Optional[str]})


class FocusFoundEvent(Event):
    """Event to be sent when a new best focus has been found, e.g. after a focus series."""

    __module__ = "pyobs.events"

    def __init__(self, focus: float, error: Optional[float] = None, filter_name: Optional[str] = None, **kwargs: Any):
        Event.__init__(self)
        self.data: DataType = {"focus": focus, "error": error, "filter_name": filter_name}

    @property
    def focus(self) -> float:
        return self.data["focus"]

    @property
    def error(self) -> Optional[float]:
        return self.data["error"]

    @property
    def filter_name(self) -> Optional[str]:
        return self.data["filter_name"]


__all__ = ["FocusFoundEvent"]



import datetime
import functools
import logging
from typing import Union, Optional, Any, Dict
import pytz
from astropy.coordinates import EarthLocation, Longitude, SkyCoord, ICRS, get_sun, AltAz
from pyobs.utils.time import Time


log = logging.getLogger(__name__)


class Environment:
    """An Environment object hold information about the location of the observatory and provides some convenience
    methode.

    An object of this type is usually never instantiated manually, but is part of a :class:`~pyobs.modules.Module`,
    automatically created from the configuration file like this::

        timezone: Africa/Johannesburg
        location:
            longitude: 20.810808
            latitude: -32.375823
            elevation: 1798.
    """

    def __init__(
        self, timezone: str = "utc", location: Union[Dict[str, Any], EarthLocation] = None, *args: Any, **kwargs: Any
    ):
        # get timezone
        self._timezone = pytz.timezone(timezone)
        log.info("Using timezone %s.", timezone)

        # get location
        self._location: Optional[EarthLocation] = None
        if location is not None:
            if isinstance(location, EarthLocation):
                # store directly
                self._location = location
            elif isinstance(location, dict):
                # dictionary?
                if (
                    "longitude" in location
                    and location["longitude"] is not None
                    and "latitude" in location
                    and location["latitude"] is not None
                    and "elevation" in location
                    and location["elevation"] is not None
                ):
                    self._location = EarthLocation(location["longitude"], location["latitude"], location["elevation"])
                else:
                    log.error("Location must be provided as dict of longitude/latitude/elevation values.")
            else:
                # nothing
                log.error("Could not initialize location.")
        if self._location is not None:
            log.info(
                "Setting location to longitude=%s, latitude=%s, and elevation=%s.",
                self._location.lon,
                self._location.lat,
                self._location.height,
            )

    @property
    def timezone(self) -> datetime.tzinfo:
        """Returns the timezone of the observatory."""
        return self._timezone

    @property
    def location(self) -> Optional[EarthLocation]:
        """Returns the location of the observatory."""
        return self._location

    @functools.lru_cache()
    def localtime(self, utc: Optional[datetime.datetime] = None) -> datetime.datetime:
        """Returns the local time at the observatory, either for a given UTC time or for now, if none is given.

        Args:
            utc: UTC to convert. Use now if none is given.

        Returns:
            Local time.
        """
        # get UTC
        if utc is None:
            utc = datetime.datetime.utcnow()
        # mark as UTC
        utc_dt = pytz.utc.localize(utc)
        # convert to local timezone
        return utc_dt.astimezone(self._timezone)

    @functools.lru_cache()
    def night_obs(self, time: Optional[Union[datetime.datetime, Time]] = None) -> datetime.date:
        """Returns the date of the night for the given night, i.e. the date of the start of the night.

        Args:
            time: Time to return night for. If none is given, current time is used.

        Returns:
            Night of observation.
        """

        # None given?
        if time is None:
            time = Time.now()

        # convert to Time
        if isinstance(time, Time):
            time = time.datetime

        # get local datetime
        if not isinstance(time, datetime.datetime):
            raise ValueError("Invalid time")
        utc_dt = pytz.utc.localize(time)
        loc_dt = utc_dt.astimezone(self._timezone)

        # get night
        if loc_dt.hour < 15:
            loc_dt += datetime.timedelta(days=-1)
        return loc_dt.date()

    @functools.lru_cache()
    def lst(self, time: Union[datetime.datetime, Time]) -> Longitude:
        """Returns the local sidereal time for a given time.

        Args:
            time: Time to convert to LST.

        Returns:
            Local sidereal time.
        """

        # no location
        if not isinstance(self._location, EarthLocation):
            raise ValueError("No location given.")

        # convert to Time
        if not isinstance(time, Time):
            time = Time(time)
        # return LST
        return time.sidereal_time("mean", longitude=self._location.lon)

    @functools.lru_cache()
    def zenith_position(self, lst: Longitude) -> SkyCoord:
        """Returns the RA/Dec position of the zenith.

        Args:
            lst: Local sidereal time to use.

        Returns:
            SkyCoord with current zenith position.
        """
        if not isinstance(self._location, EarthLocation):
            raise ValueError("No location given.")
        return SkyCoord(lst, self._location.lat, frame=ICRS)

    def now(self) -> Time:
        """Returns current time."""
        return Time.now()

    def to_altaz(self, radec: SkyCoord, time: Optional[Time] = None) -> SkyCoord:
        """Converts a given set of RA/Dec to Alt/Az for the current location at a given time.

        Args:
            radec: RA/Dec coordinates to convert.
            time: Time to use, or none for now.

        Returns:
            Alt/Az coordinates for given RA/Dec.
        """
        if time is None:
            time = Time.now()
        return radec.transform_to(AltAz(obstime=time, location=self.location))

    def to_radec(self, altaz: SkyCoord, time: Optional[Time] = None) -> SkyCoord:
        """Converts a given set of Alt/Az to RA/Dec for the current location at a given time.

        Args:
            altaz: Alt/Az coordinates to convert.
            time: Time to use, or none for now.

        Returns:
            RA/Dec coordinates for given Alt/Az.
        """
        if time is None:
            time = Time.now()
        altaz.location = self.location
        altaz.obstime = time
        return altaz.icrs

    @functools.lru_cache()
    def sun(self, time: Time, altaz: bool = True) -> SkyCoord:
        """Returns the position of the sun, either as RA/Dec or Alt/Az for the given time.

        Args:
            time: Time to calculate position for.
            altaz: If True, Alt/Az is returned, otherwise RA/Dec.

        Returns:
            Coordinates of sun.
        """

        # alt/az or ra/dec?
        if altaz:
            return self.to_altaz(get_sun(time), time)
        else:
            return get_sun(time)


__all__ = ["Environment"]



import asyncio
import logging
from typing import Tuple, Any, Dict, List, Optional
import aiohttp
import urllib.parse
import astropy.units as u

from pyobs.utils.enums import WeatherSensors
from pyobs.utils.time import Time
from pyobs.events import BadWeatherEvent, GoodWeatherEvent
from pyobs.interfaces import IWeather, IFitsHeaderBefore
from pyobs.modules import Module


log = logging.getLogger(__name__)


FITS_HEADERS = {
    WeatherSensors.TEMPERATURE: ("WS-TEMP", "Ambient temperature average during exposure, C", float),
    WeatherSensors.HUMIDITY: ("WS-HUMID", "Ambient rel. humidity average, %", float),
    WeatherSensors.PRESSURE: ("WS-PRESS", "Average atmospheric pressure, hPa", float),
    WeatherSensors.WINDDIR: ("WS-AZ", "Average wind direction, not corrected for overlap, deg", float),
    WeatherSensors.WINDSPEED: ("WS-WIND", "Ambient average wind speed, km/h", float),
    WeatherSensors.RAIN: ("WS-PREC", "Ambient precipitation [0/1]", bool),
    WeatherSensors.SKYTEMP: ("WS-SKY", "Average sky temperature, C", float),
    WeatherSensors.DEWPOINT: ("WS-TDEW", "Ambient dewpoint average during expsoure, C", float),
    WeatherSensors.PARTICLES: ("WS-DUST", "Average particle count during exposure, ppcm", float),
}


class Weather(Module, IWeather, IFitsHeaderBefore):
    """Connection to pyobs-weather."""

    __module__ = "pyobs.modules.weather"

    def __init__(self, url: str, system_init_time: int = 300, **kwargs: Any):
        """Initialize a new pyobs-weather connector.

        Args:
            url: URL to weather station
            system_init_time: Time in seconds the full system needs to initialize
        """
        Module.__init__(self, **kwargs)

        # store and create session
        self._system_init_time = system_init_time
        self._url = url

        # whether module is active, i.e. if None, weather is always good
        self._active = True

        # current status
        self._is_good: Optional[bool] = None

        # whole status
        self._status: Dict[str, Any] = {}

        # add thread func
        self.add_background_task(self._update, True)

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # subscribe to events
        if self.comm:
            await self.comm.register_event(BadWeatherEvent)
            await self.comm.register_event(GoodWeatherEvent)

    async def start(self, **kwargs: Any) -> None:
        """Starts a service."""

        # did status change and weather is now bad?
        if not self._active and not self._is_good:
            # send event!
            await self.comm.send_event(BadWeatherEvent())

        # activate
        self._active = True

    async def stop(self, **kwargs: Any) -> None:
        """Stops a service."""
        self._active = False

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether a service is running."""
        return self._active

    async def _update(self) -> None:
        """Update weather info."""

        # loop forever
        while True:
            # new is_good status
            is_good: Optional[bool] = None
            error = False

            try:
                # fetch status
                url = urllib.parse.urljoin(self._url, "api/current/")
                async with aiohttp.ClientSession() as session:
                    async with session.get(url, timeout=5) as response:
                        if response.status != 200:
                            raise ValueError("Could not connect to weather station.")
                        status = await response.json()

                # to json
                if "good" not in status:
                    raise ValueError("Good parameter not found in response from weather station.")

                # store it
                is_good = status["good"]
                self._status = status

            except Exception as e:
                # on error, we're always bad
                log.error("Request failed: %s", e)
                is_good = False
                error = True

            # did status change?
            if is_good != self._is_good:
                # only send changes, if active
                if self._active:
                    # did it change to good or bad?
                    if is_good:
                        log.info("Weather is now good.")
                        eta = Time.now() + self._system_init_time * u.second
                        await self.comm.send_event(GoodWeatherEvent(eta=eta))
                    else:
                        log.info("Weather is now bad.")
                        await self.comm.send_event(BadWeatherEvent())

                # store new state
                self._is_good = is_good

            # sleep a little
            await asyncio.sleep(60 if error else 5)

    async def get_weather_status(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns status of object in form of a dictionary. See other interfaces for details."""
        raise NotImplementedError

    async def is_weather_good(self, **kwargs: Any) -> bool:
        """Whether the weather is good to observe."""

        # if not active, weather is always good
        if not self._active:
            return True

        # otherwise it depends on the is_good flag
        return False if self._is_good is None else self._is_good

    async def get_current_weather(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns current weather.

        Returns:
            Dictionary containing entries for time, good, and sensor, with the latter being another dictionary
            with sensor information, which contain a value and a good flag.
        """
        return self._status

    async def get_sensor_value(self, station: str, sensor: WeatherSensors, **kwargs: Any) -> Tuple[str, float]:
        """Return value for given sensor.

        Args:
            station: Name of weather station to get value from.
            sensor: Name of sensor to get value from.

        Returns:
            Tuple of current value of given sensor or None and time of measurement or None.
        """

        # do request
        url = urllib.parse.urljoin(self._url, "api/stations/%s/%s/" % (station, sensor.value))
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=5) as response:
                if response.status != 200:
                    raise ValueError("Could not connect to weather station.")
                status = await response.json()

        # to json
        if "time" not in status or "value" not in status:
            raise ValueError("Time and/or value parameters not found in response from weather station.")

        # return time and value
        return status["time"], status["value"]

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # copy status
        status = dict(self._status)

        # got sensors?
        if "sensors" not in status:
            log.error("No sensor data found in status.")
            return {}
        sensors = status["sensors"]

        # loop sensor types
        header = {}
        for sensor_type in WeatherSensors:
            # got a value for this type?
            if sensor_type.value in sensors:
                # get value
                if "value" not in sensors[sensor_type.value]:
                    continue
                value = sensors[sensor_type.value]["value"]

                # get header keyword, comment and data type
                key, comment, dtype = FITS_HEADERS[sensor_type]

                # set it
                header[key] = (None if value is None else dtype(value), comment)

        # finished
        return header


__all__ = ["Weather"]



from __future__ import annotations

import asyncio
import threading
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u
from typing import Tuple, List, Optional, Any, TYPE_CHECKING
import random

from pyobs.object import Object
from pyobs.utils.enums import MotionStatus
from pyobs.utils.parallel import event_wait

if TYPE_CHECKING:
    from pyobs.utils.simulation import SimWorld


class SimTelescope(Object):
    """A simulated telescope on an equitorial mount."""

    __module__ = "pyobs.utils.simulation"

    def __init__(
        self,
        world: "SimWorld",
        position: Optional[Tuple[float, float]] = None,
        offsets: Optional[Tuple[float, float]] = None,
        pointing_offset: Optional[Tuple[float, float]] = None,
        move_accuracy: float = 2.0,
        speed: float = 20.0,
        focus: float = 50,
        filters: Optional[List[str]] = None,
        filter_name: str = "clear",
        drift: Optional[Tuple[float, float]] = None,
        focal_length: float = 5000.0,
        **kwargs: Any,
    ):
        """Initializes new telescope.

        Args:
            world: World object.
            position: RA/Dec tuple with position of telescope in degrees.
            offsets: RA/Dec offsets of telescope in arcsecs.
            pointing_offset: Pointing offset in RA/Dec in arcsecs.
            move_accuracy: Accuracy of movements in RA/Dec, i.e. random error after any movement [arcsec].
            speed: Speed of telescope in deg/sec.
            focus: Telescope focus.
            filters: List of filters.
            filter_name: Current filter.
            drift: RA/Dec drift of telescope in arcsec/sec.
            focal_length: Focal length of telescope in mm.
        """
        Object.__init__(self, **kwargs)

        # store
        self.world = world
        self.status = MotionStatus.IDLE
        self.status_callback = None

        # init
        self._position = (
            SkyCoord(0.0 * u.deg, 0.0 * u.deg, frame="icrs")
            if position is None
            else SkyCoord(position[0] * u.deg, position[1] * u.deg, frame="icrs")
        )
        self._offsets = (0.0, 0.0) if offsets is None else offsets
        self.pointing_offset = (20.0, 2.0) if pointing_offset is None else pointing_offset
        self.move_accuracy = (1, 1) if move_accuracy is None else move_accuracy
        self.speed = speed  # telescope speed in deg/sec
        self.focus = focus
        self.filters = ["clear", "B", "V", "R"] if filters is None else filters
        self.filter_name = filter_name
        self.drift = (0.0, 0.0) if drift is None else drift  # arcsec/sec in RA/Dec
        self.focal_length = focal_length

        # private stuff
        self._drift = (0.0, 0.0)
        self._dest_coords = None

        # locks
        self._pos_lock = threading.RLock()

        # threads
        self.add_background_task(self._move_task)

    @property
    def position(self) -> SkyCoord:
        return self._position

    @property
    def offsets(self) -> Tuple[float, float]:
        return self._offsets

    def _change_motion_status(self, status: MotionStatus) -> None:
        """Change the current motion status.

        Args:
            status: New motion status
        """

        # call callback
        if self.status_callback is not None and status != self.status:
            self.status_callback(status)

        # set it
        self.status = status

    @property
    def real_pos(self) -> SkyCoord:
        # calculate offsets
        dra = (self._offsets[0] * u.deg + self._drift[0] * u.arcsec) / np.cos(np.radians(self._position.dec.degree))
        ddec = self._offsets[1] * u.deg + self._drift[1] * u.arcsec

        # return position
        with self._pos_lock:
            return SkyCoord(ra=self._position.ra + dra, dec=self._position.dec + ddec, frame="icrs")

    def move_ra_dec(self, coords: SkyCoord) -> None:
        """Move telescope to given RA/Dec position.

        Args:
            coords: Destination coordinates.
        """

        # change status
        self._change_motion_status(MotionStatus.SLEWING)

        # calculate random RA/Dec offsets
        acc = self.move_accuracy / 3600.0
        ra = random.gauss(coords.ra.degree, acc / np.cos(np.radians(coords.dec.degree))) * u.deg
        dec = random.gauss(coords.dec.degree, acc) * u.deg

        # set coordinates
        self._dest_coords = SkyCoord(ra=ra, dec=dec, frame="icrs")

    def set_offsets(self, dra: float, ddec: float) -> None:
        """Move RA/Dec offsets.

        Args:
            dra: RA offset [deg]
            ddec: Dec offset [deg]
        """

        # calculate random RA/Dec offsets
        acc = self.move_accuracy / 3600.0
        ra, dec = random.gauss(dra, acc), random.gauss(ddec, acc)

        # set offsets
        self._offsets = (ra, dec)

    async def _move_task(self) -> None:
        """Move the telescope over time."""

        # run until closed
        while True:

            # do we have destination coordinates?
            if self._dest_coords is not None:
                # calculate moving vector
                vra = (self._dest_coords.ra.degree - self._position.ra.degree) * np.cos(
                    np.radians(self._position.dec.degree)
                )
                vdec = self._dest_coords.dec.degree - self._position.dec.degree

                # get direction
                length = np.sqrt(vra ** 2 + vdec ** 2)

                # do we reach target?
                if length < self.speed:
                    # set it
                    with self._pos_lock:
                        # set position and reset destination
                        self._change_motion_status(MotionStatus.TRACKING)
                        self._position = self._dest_coords
                        self._dest_coords = None

                        # set some random drift around the pointing error
                        self._drift = (
                            random.gauss(self.pointing_offset[0], self.pointing_offset[0] / 10.0),
                            random.gauss(self.pointing_offset[1], self.pointing_offset[1] / 10.0),
                        )

                else:
                    # norm vector and get movement
                    dra = vra / length * self.speed / np.cos(np.radians(self._position.dec.degree)) * u.deg
                    ddec = vdec / length * self.speed * u.deg

                    # apply it
                    with self._pos_lock:
                        self._change_motion_status(MotionStatus.SLEWING)
                        self._position = SkyCoord(
                            ra=self._position.ra + dra, dec=self._position.dec + ddec, frame="icrs"
                        )

            else:
                # no movement, just drift
                # calculate constant drift
                drift_ra = random.gauss(self.drift[0], self.drift[0] / 10.0)
                drift_dec = random.gauss(self.drift[1], self.drift[1] / 10.0)

                # and apply it
                with self._pos_lock:
                    self._drift = (self._drift[0] + drift_ra, self._drift[1] + drift_dec)

            # sleep a second
            await asyncio.sleep(1)


__all__ = ["SimTelescope"]



from pyobs.interfaces import ITelescope
from pyobs.object import Object
from pyobs.utils.parallel import Future


class SkyFlatsBasePointing(Object):
    """Base class for flat poinings."""

    __module__ = "pyobs.utils.skyflats.pointing"

    async def __call__(self, telescope: ITelescope) -> None:
        """Move telescope.

        Args:
            telescope: Telescope to use.

        Returns:
            Future for the movement call.
        """
        raise NotImplementedError

    async def reset(self) -> None:
        """Reset pointing."""
        pass


__all__ = ["SkyFlatsBasePointing"]



from typing import Optional, Any, Dict
from typing_extensions import TypedDict

from pyobs.utils.time import Time
from pyobs.events.event import Event


DataType = TypedDict("DataType", {"name": str, "id": Any, "eta": Optional[str]})


class TaskStartedEvent(Event):
    """Event to be sent when a task has started."""

    __module__ = "pyobs.events"

    def __init__(self, name: str, id: Any, eta: Optional[Time] = None, **kwargs: Any):
        """Initializes a new task started event.

        Args:
            name: Name of task that just started
            id: Unique identifier for task
            eta: Predicted ETA for when the task will finish
        """
        Event.__init__(self)
        self.data: DataType = {"name": name, "id": id, "eta": None if eta is None else eta.isot}

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Event:
        # get name
        if "name" not in d or not isinstance(d["name"], str):
            raise ValueError("Invalid type for name.")
        name: str = d["name"]

        # get id
        if "id" not in d:
            raise ValueError("Invalid type for id.")
        id: Any = d["id"]

        # get eta
        eta: Optional[Time] = None
        if "eta" in d and isinstance(d["eta"], str):
            eta = Time(d["eta"])

        # return object
        return TaskStartedEvent(name=name, id=id, eta=eta)

    @property
    def name(self) -> str:
        return self.data["name"]

    @property
    def id(self) -> Any:
        return self.data["id"]

    @property
    def eta(self) -> Optional[Time]:
        return Time(self.data["eta"]) if self.data["eta"] is not None else None


__all__ = ["TaskStartedEvent"]



import logging
from abc import ABCMeta, abstractmethod

from astropy.wcs import WCS

from pyobs.images import Image
from pyobs.images.processor import ImageProcessor


log = logging.getLogger(__name__)


class Offsets(ImageProcessor, metaclass=ABCMeta):
    """Base class for determining offsets."""

    __module__ = "pyobs.images.processors.offsets"

    @abstractmethod
    async def __call__(self, image: Image) -> Image:
        """Processes an image and sets x/y pixel offset to reference in meta data.

        Args:
            image: Image to process.

        Returns:
            Original image.

        Raises:
            ValueError: If offset could not be found.
        """
        raise NotImplementedError

    @staticmethod
    def on_sky_distance(image: Image) -> float:
        """Calculate on sky distance for offset calculated by any of the derived classes in degrees.

        Args:
            image: Image to analyse. Needs CRPIX1/2 entries in header and 'offsets' in meta.

        Returns:
            On-sky offset of offset in degrees.
        """

        # get central position and offset
        center = image.header["CRPIX1"], image.header["CRPIX2"]
        offsets = image.meta["offsets"]

        # get WCS
        wcs = WCS(image.header)

        # get RA/Dec at center and at center+offsets
        center_coord = wcs.pixel_to_world(*center)
        offset_coord = wcs.pixel_to_world(*tuple(map(sum, zip(center, offsets))))

        # calculate distance and return it
        return float(center_coord.separation(offset_coord).value)


__all__ = ["Offsets"]
